# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

import pytest
import numpy as np
from numpy.testing import assert_allclose
import pickle

from pathlib import Path

import MDAnalysis as mda
from MDAnalysis import units
from MDAnalysis.auxiliary.EDR import HAS_PYEDR

from MDAnalysisTests.datafiles import (
    AUX_EDR,
    AUX_EDR_TPR,
    AUX_EDR_XTC,
    AUX_EDR_RAW,
    AUX_EDR_SINGLE_FRAME,
)
from MDAnalysisTests.auxiliary.base import (
    BaseAuxReaderTest,
    BaseAuxReference,
    assert_auxstep_equal,
)


def read_raw_data_file(step):
    """parses raw edr data (plain text) and returns dictionary with auxdata for
    step"""
    with open(AUX_EDR_RAW) as f:
        rawdata = f.readlines()
    n_entries = 52  # number of aux terms per step
    stepdata = rawdata[step * n_entries : (step + 1) * n_entries]
    aux_dict = {}
    edr_units = {}
    for line in stepdata:
        aux_dict[line.split(",")[0]] = float(line.split(",")[1])
        edr_units[line.split(",")[0]] = line.split(",")[2].strip()
    return aux_dict, edr_units


def get_auxstep_data(step):
    return read_raw_data_file(step)[0]


def get_edr_unit_dict(step):
    return read_raw_data_file(step)[1]


@pytest.mark.skipif(HAS_PYEDR, reason="pyedr present")
def test_pyedr_not_present_raises():
    with pytest.raises(ImportError, match="please install pyedr"):
        aux = mda.auxiliary.EDR.EDRReader(AUX_EDR)


@pytest.mark.skipif(not HAS_PYEDR, reason="pyedr not installed")
class EDRReference(BaseAuxReference):
    """
    Class to hold reference values for :class:`TestEDRReader` to test against.

    Because of the binary file format, it is not possible to create a test file
    of the format that is assumed for :class:`BaseAuxReference`. As such, some
    adaptations were necessary for the EDRReader.
    """

    def __init__(self):
        super(EDRReference, self).__init__()
        self.testdata = AUX_EDR
        self.reader = mda.auxiliary.EDR.EDRReader
        self.n_steps = 4
        self.dt = 0.02
        self.description = {
            "dt": self.dt,
            "represent_ts_as": "closest",
            "initial_time": self.initial_time,
            "time_selector": "Time",
            "data_selector": None,
            "constant_dt": True,
            "cutoff": None,
            "auxname": self.name,
        }

        def reference_auxstep(i):
            # create a reference AuxStep for step i
            t_init = self.initial_time
            auxstep = mda.auxiliary.EDR.EDRStep(
                dt=self.dt, initial_time=t_init
            )
            auxstep.step = i
            auxstep._data = get_auxstep_data(i)
            return auxstep

        self.auxsteps = [reference_auxstep(i) for i in range(self.n_steps)]

        # add the auxdata and format for .edr to the reference description
        self.description["auxdata"] = Path(self.testdata).resolve()
        self.description["format"] = self.reader.format

        # for testing the selection of data/time
        self.time_selector = "Time"
        self.select_time_ref = [
            step._data[self.time_selector] for step in self.auxsteps
        ]
        self.data_selector = "Bond"  # selects all data
        self.select_data_ref = [step._data["Bond"] for step in self.auxsteps]

        # testing __getitem__ with slice and list. Should allow us to iterate
        # through the specified auxiliary steps...
        self.iter_list = [0, -2]
        self.iter_list_auxsteps = [self.auxsteps[0], self.auxsteps[2]]
        self.iter_slice = slice(None, None, 2)  # every second step
        self.iter_slice_auxsteps = [self.auxsteps[0], self.auxsteps[2]]

        def reference_timestep(dt=0.02, offset=0):
            # return a trajectory timestep with specified dt, offset + move to
            # frame 1; for use in auxiliary reading of different timesteps
            ts = mda.coordinates.base.Timestep(0, dt=dt, time_offset=offset)
            ts.frame = 1
            return ts

        self.bond_step1_val = 1426.2252197265625
        self.bond_step2_val = 1482.0098876953125
        self.bond_step1_2_avg = (self.bond_step1_val + self.bond_step2_val) / 2
        # test reading a timestep with lower frequency. Auxiliary steps with
        # times between [0.02 ps, 0.06 ps) will be assigned to this timestep,
        # i.e. step 1 (0.02 ps) and step 2 (0.04 ps).
        self.lower_freq_ts = reference_timestep(dt=0.04, offset=0)
        # 'closest' representative value will match step 2
        # Pick value for "Bond" to check against.
        self.lowf_closest_rep = self.bond_step2_val
        # 'average' representative value
        self.lowf_avg_time = 0.03
        self.lowf_average_rep = [self.lowf_avg_time, self.bond_step1_2_avg]

        # test reading a timestep with higher frequency. Auxiliart steps with
        # times between [0.25ps, 0.75ps) will be assigned to this timestep,
        # i.e. no auxiliary steps
        self.higher_freq_ts = reference_timestep(dt=0.01, offset=0)
        self.highf_rep = self.format_data(np.nan)

        # test reading a timestep that is offset. Auxiliary steps with
        # times between [0.015ps, 0.035ps) will be assigned to this timestep,
        # i.e step 1 (0.02 ps)
        self.offset_ts = reference_timestep(dt=0.02, offset=0.005)
        # 'closest' representative value will match step 1 data
        self.offset_closest_rep = self.bond_step1_val

        # testing cutoff for representative values
        self.cutoff = 0
        # for 'average': use low frequenct timestep, only step 2 within 0ps
        # cutoff
        self.lowf_cutoff_average_rep = self.bond_step2_val
        # for 'closest': use offset timestep; no timestep within 0ps cutoff
        # changed here to match requirements for EDR data
        self.offset_cutoff_closest_rep = np.array(np.nan)

        # for testing EDRReader.get_data()
        self.times = np.array([0.0, 0.02, 0.04, 0.06])
        self.bonds = np.array(
            [1374.82324219, 1426.22521973, 1482.0098877, 1470.33752441]
        )
        self.angles = np.array(
            [3764.52734375, 3752.83032227, 3731.59179688, 3683.40942383]
        )


@pytest.mark.skipif(not HAS_PYEDR, reason="pyedr not installed")
class TestEDRReader(BaseAuxReaderTest):
    """Class to conduct tests for the auxiliary EDRReader

    Normally, it would be desirable to use the tests from
    :class:`BaseAuxReaderTest`, but this is not possible for some of the tests
    there because of the differences between :class:`BaseAuxReference` and
    :class:`EDRReference` which are ultimately due to the different file
    format. Therefore, some tests had to be overridden here. New tests for EDR-
    specific functionality were also added.
    """

    @staticmethod
    @pytest.fixture
    def ref():
        return EDRReference()

    @staticmethod
    @pytest.fixture
    def ref_universe(ref):
        u = mda.Universe(AUX_EDR_TPR, AUX_EDR_XTC)
        # TODO: Change order of aux_spec and auxdata for 3.0 release, cf. Issue #3811
        u.trajectory.add_auxiliary({"test": "Bond"}, ref.testdata)
        return u

    @staticmethod
    @pytest.fixture
    def reader(ref):
        reader = ref.reader(
            ref.testdata,
            initial_time=ref.initial_time,
            dt=ref.dt,
            auxname=ref.name,
            time_selector="Time",
            data_selector=None,
        )
        ref_units = get_edr_unit_dict(0)
        if reader.unit_dict != ref_units:
            for term, ref_unit in ref_units.items():
                data = reader.data_dict[term]
                reader_unit = reader.unit_dict[term]
                try:
                    reader.data_dict[term] = units.convert(
                        data, reader_unit, ref_unit
                    )
                except ValueError:
                    continue  # some units not supported yet
        reader.rewind()
        return reader

    def test_time_non_constant_dt(self, reader):
        reader.constant_dt = False
        reader.time_selector = None
        with pytest.raises(
            ValueError,
            match="If dt is not constant, "
            "must have a valid time "
            "selector",
        ):
            reader.time

    def test_iterate_through_trajectory(self, ref, ref_universe):
        # check the representative values of aux for each frame are as expected
        # trajectory here has same dt, offset; so there's a direct
        # correspondence between frames and steps
        for i, ts in enumerate(ref_universe.trajectory):
            assert_allclose(ts.aux.test, ref.auxsteps[i].data["Bond"])

    def test_iterate_as_auxiliary_from_trajectory(self, ref, ref_universe):
        # check representative values of aux for each frame are as expected
        # trajectory here has same dt, offset, so there's a direct
        # correspondence between frames and steps, and iter_as_aux will run
        # through all frames
        for i, ts in enumerate(ref_universe.trajectory.iter_as_aux("test")):
            assert_allclose(ts.aux.test, ref.auxsteps[i].data["Bond"])

    def test_step_to_frame_time_diff(self, reader, ref):
        # Timestep is 0.002 (10 %) longer than auxiliary data
        ts = mda.coordinates.base.Timestep(0, dt=ref.dt + 0.002)

        # Test all 4 frames
        for idx in range(4):

            frame, time_diff = reader.step_to_frame(
                idx, ts, return_time_diff=True
            )

            assert frame == idx
            assert_allclose(time_diff, idx * 0.002)

    def test_read_lower_freq_timestep(self, ref, reader):
        # test reading a timestep with lower frequency
        ts = ref.lower_freq_ts
        reader.update_ts(ts)
        # check the value set in ts is as we expect
        assert_allclose(
            ts.aux.test["Bond"],
            ref.lowf_closest_rep,
            err_msg="Representative value in ts.aux " "does not match",
        )

    def test_read_higher_freq_timestep(self, ref, reader):
        # try reading a timestep with higher frequency
        ts = ref.higher_freq_ts
        reader.update_ts(ts)
        assert_allclose(
            ts.aux.test,
            ref.highf_rep,
            err_msg="Representative value in ts.aux " "does not match",
        )

    def test_read_offset_timestep(self, ref, reader):
        # try reading a timestep offset from auxiliary
        ts = ref.offset_ts
        reader.update_ts(ts)
        assert_allclose(
            ts.aux.test["Bond"],
            ref.offset_closest_rep,
            err_msg="Representative value in ts.aux " "does not match",
        )

    def test_represent_as_average(self, ref, reader):
        # test the 'average' option for 'represent_ts_as'
        # reset the represent method to 'average'...
        reader.represent_ts_as = "average"
        # read timestep; use the low freq timestep
        ts = ref.lower_freq_ts
        reader.update_ts(ts)
        # check the representative value set in ts is as expected
        test_value = [ts.aux.test["Time"], ts.aux.test["Bond"]]
        assert_allclose(
            test_value,
            ref.lowf_average_rep,
            err_msg="Representative value does not match when "
            "using with option 'average'",
        )

    def test_represent_as_average_with_cutoff(self, ref, reader):
        # test the 'represent_ts_as' 'average' option when we have a cutoff set
        # set the cutoff...
        reader.cutoff = ref.cutoff
        # read timestep; use the low frequency timestep
        ts = ref.lower_freq_ts
        reader.update_ts(ts)
        # check representative value set in ts is as expected
        assert_allclose(
            ts.aux.test["Bond"],
            ref.lowf_cutoff_average_rep,
            err_msg="Representative value does not match when "
            "applying cutoff",
        )

    # TODO: Change order of aux_spec and auxdata for 3.0 release, cf. Issue #3811
    def test_add_all_terms_from_file(self, ref, ref_universe):
        ref_universe.trajectory.add_auxiliary(auxdata=ref.testdata)
        # adding "test" manually to match above addition of test term
        ref_terms = ["test"] + [key for key in get_auxstep_data(0).keys()]
        terms = [key for key in ref_universe.trajectory._auxs]
        assert ref_terms == terms

    # TODO: Change order of aux_spec and auxdata for 3.0 release, cf. Issue #3811
    def test_add_all_terms_from_reader(self, ref_universe, reader):
        ref_universe.trajectory.add_auxiliary(auxdata=reader)
        ref_terms = ["test"] + [key for key in get_auxstep_data(0).keys()]
        terms = [key for key in ref_universe.trajectory._auxs]
        assert ref_terms == terms

    # TODO: Change order of aux_spec and auxdata for 3.0 release, cf. Issue #3811
    def test_add_term_list_custom_names_from_file(self, ref, ref_universe):
        ref_universe.trajectory.add_auxiliary(
            {"bond": "Bond", "temp": "Temperature"}, ref.testdata
        )
        ref_dict = get_auxstep_data(0)
        assert ref_universe.trajectory.ts.aux.bond == ref_dict["Bond"]
        assert ref_universe.trajectory.ts.aux.temp == ref_dict["Temperature"]

    # TODO: Change order of aux_spec and auxdata for 3.0 release, cf. Issue #3811
    def test_add_term_list_custom_names_from_reader(
        self, ref_universe, reader
    ):
        ref_universe.trajectory.add_auxiliary(
            {"bond": "Bond", "temp": "Temperature"}, reader
        )
        ref_dict = get_auxstep_data(0)
        assert ref_universe.trajectory.ts.aux.bond == ref_dict["Bond"]
        assert ref_universe.trajectory.ts.aux.temp == ref_dict["Temperature"]

    # TODO: Change order of aux_spec and auxdata for 3.0 release, cf. Issue #3811
    def test_raise_error_if_auxname_already_assigned(
        self, ref_universe, reader
    ):
        with pytest.raises(ValueError, match="Auxiliary data with name"):
            ref_universe.trajectory.add_auxiliary("test", reader, "Bond")

    # TODO: Change order of aux_spec and auxdata for 3.0 release, cf. Issue #3811
    def test_add_single_term_custom_name_from_file(self, ref, ref_universe):
        ref_universe.trajectory.add_auxiliary(
            {"temp": "Temperature"}, ref.testdata
        )
        ref_dict = get_auxstep_data(0)
        assert ref_universe.trajectory.ts.aux.temp == ref_dict["Temperature"]

    # TODO: Change order of aux_spec and auxdata for 3.0 release, cf. Issue #3811
    def test_add_single_term_custom_name_from_reader(
        self, ref_universe, reader
    ):
        ref_universe.trajectory.add_auxiliary({"temp": "Temperature"}, reader)
        ref_dict = get_auxstep_data(0)
        assert ref_universe.trajectory.ts.aux.temp == ref_dict["Temperature"]

    # TODO: Change order of aux_spec and auxdata for 3.0 release, cf. Issue #3811
    def test_terms_update_on_iter(self, ref_universe, reader):
        ref_universe.trajectory.add_auxiliary(
            {"bond": "Bond", "temp": "Temperature"}, reader
        )
        ref_dict = get_auxstep_data(0)
        assert ref_universe.trajectory.ts.aux.bond == ref_dict["Bond"]
        assert ref_universe.trajectory.ts.aux.temp == ref_dict["Temperature"]
        ref_dict = get_auxstep_data(1)
        ref_universe.trajectory.next()
        assert ref_universe.trajectory.ts.aux.bond == ref_dict["Bond"]
        assert ref_universe.trajectory.ts.aux.temp == ref_dict["Temperature"]

    # TODO: Change order of aux_spec and auxdata for 3.0 release, cf. Issue #3811
    def test_invalid_data_selector(self, ref, ref_universe):
        with pytest.raises(KeyError, match="'Nonsense' is not a key"):
            ref_universe.trajectory.add_auxiliary(
                {"something": "Nonsense"}, AUX_EDR
            )

    def test_read_all_times(self, reader):
        all_times_expected = np.array([0.0, 0.02, 0.04, 0.06])
        assert_allclose(all_times_expected, reader.read_all_times())

    def test_get_data_from_string(self, ref, reader):
        returned = reader.get_data("Bond")
        assert isinstance(returned, dict)
        assert_allclose(ref.times, returned["Time"])
        assert_allclose(ref.bonds, returned["Bond"])

    def test_get_data_from_list(self, ref, reader):
        returned = reader.get_data(["Bond", "Angle"])
        assert isinstance(returned, dict)
        assert_allclose(ref.times, returned["Time"])
        assert_allclose(ref.bonds, returned["Bond"])
        assert_allclose(ref.angles, returned["Angle"])

    def test_get_data_everything(self, ref, reader):
        returned = reader.get_data()
        returned_asterisk = reader.get_data()
        assert returned.keys() == returned_asterisk.keys()
        ref_terms = [key for key in get_auxstep_data(0).keys()]
        assert ref_terms == reader.terms
        assert_allclose(ref.bonds, returned["Bond"])

    @pytest.mark.parametrize(
        "get_data_input",
        (42, "Not a valid term", ["Bond", "Not a valid term"]),
    )
    def test_get_data_invalid_selections(self, reader, get_data_input):
        with pytest.raises(KeyError, match="data selector"):
            reader.get_data(get_data_input)

    # TODO: Change order of aux_spec and auxdata for 3.0 release, cf. Issue #3811
    def test_warning_when_space_in_aux_spec(self, ref_universe, reader):
        with pytest.warns(UserWarning, match="Auxiliary name"):
            ref_universe.trajectory.add_auxiliary(
                {"Pres. DC": "Pres. DC"}, reader
            )

    # TODO: Change order of aux_spec and auxdata for 3.0 release, cf. Issue #3811
    def test_warn_too_much_memory_usage(self, ref_universe, reader):
        with pytest.warns(
            UserWarning,
            match="AuxReader: memory usage "
            "warning! Auxiliary data takes up 3[0-9.]*e-06 GB of"
            r" memory \(Warning limit: 1e-08 GB\)",
        ):
            ref_universe.trajectory.add_auxiliary(
                {"temp": "Temperature"}, reader, memory_limit=10
            )

    def test_auxreader_picklable(self, reader):
        new_reader = pickle.loads(pickle.dumps(reader))
        for step_index, auxstep in enumerate(reader):
            assert_auxstep_equal(new_reader[step_index], auxstep)

    def test_units_are_converted_by_EDRReader(self, reader):
        original_units = get_edr_unit_dict(0)
        reader_units = reader.unit_dict
        # so far, lengths and speeds are converted
        for term in ["Box-X", "Box-Vel-XX"]:
            assert original_units[term] != reader_units[term]

    # TODO: Change order of aux_spec and auxdata for 3.0 release, cf. Issue #3811
    def test_warning_when_unknown_unit(self, ref_universe, reader):
        with pytest.warns(UserWarning, match="Could not find"):
            ref_universe.trajectory.add_auxiliary(
                {"temp": "Temperature"}, reader
            )

    def test_unit_conversion_is_optional(self, ref):
        reader = ref.reader(
            ref.testdata,
            initial_time=ref.initial_time,
            dt=ref.dt,
            auxname=ref.name,
            time_selector="Time",
            data_selector=None,
            convert_units=False,
        )
        ref_units = get_edr_unit_dict(0)
        # The units from AUX_EDR match the ones from the reference
        # data file AUX_EDR_RAW. If the EDRReader does not convert the units on
        # reading the file, then the two unit dictionaries should be identical.
        assert reader.unit_dict == ref_units


@pytest.mark.skipif(not HAS_PYEDR, reason="pyedr not installed")
def test_single_frame_input_file():
    """Previously, EDRReader could not handle EDR input files with only one
    frame. See Issue #3999."""
    reader = mda.auxiliary.EDR.EDRReader(
        AUX_EDR_SINGLE_FRAME, convert_units=False
    )
    ref_dict = get_auxstep_data(0)
    reader_data_dict = reader.auxstep.data
    assert ref_dict == reader_data_dict



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
from io import StringIO

import MDAnalysis as mda
import numpy as np
import pytest
from MDAnalysis import NoDataError
from MDAnalysis.guesser import tables
from MDAnalysis.topology.PDBParser import PDBParser
from numpy.testing import assert_allclose, assert_equal

from MDAnalysisTests.datafiles import (
    PDB,
    PDB_HOLE,
    PDB_chainidnewres,
    PDB_charges,
    PDB_conect,
    PDB_conect2TER,
    PDB_elements,
    PDB_helix,
    PDB_sameresid_diffresname,
    PDB_singleconect,
    PDB_small,
)
from MDAnalysisTests.topology.base import ParserBase

_PDBPARSER = mda.topology.PDBParser.PDBParser

hybrid36 = [
    ("A0000", 100000),
    ("MEGAN", 20929695),
    ("J0NNY", 15247214),
    ("DREW6", 6417862),
    ("ST3V3", 31691119),
    ("ADA8M", 719798),
    ("a0000", 43770016),
    ("megan", 64599711),
    ("j0nny", 58917230),
    ("drew6", 50087878),
    ("st3v3", 75361135),
    ("ada8m", 44389814),
    ("    6", 6),
    ("   24", 24),
    ("  645", 645),
    (" 4951", 4951),
    ("10267", 10267),
]


@pytest.mark.parametrize("hybrid, integer", hybrid36)
def test_hy36decode(hybrid, integer):
    assert mda.topology.PDBParser.hy36decode(5, hybrid) == integer


class PDBBase(ParserBase):
    expected_attrs = [
        "ids",
        "names",
        "record_types",
        "resids",
        "resnames",
        "altLocs",
        "icodes",
        "occupancies",
        "tempfactors",
        "chainIDs",
    ]
    guessed_attrs = ["types", "masses"]


class TestPDBParser(PDBBase):
    """This one has neither chainids or segids"""

    parser = mda.topology.PDBParser.PDBParser
    ref_filename = PDB
    expected_n_atoms = 47681
    expected_n_residues = 11302
    expected_n_segments = 1


class TestPDBParserSegids(PDBBase):
    """Has segids"""

    parser = mda.topology.PDBParser.PDBParser
    ref_filename = PDB_small
    expected_n_atoms = 3341
    expected_n_residues = 214
    expected_n_segments = 1


class TestPDBConect(object):
    """Testing PDB topology parsing (PDB)"""

    def test_conect_parser(self):
        lines = (
            "CONECT1233212331",
            "CONECT123331233112334",
            "CONECT123341233312335",
            "CONECT123351233412336",
            "CONECT12336123271233012335",
            "CONECT12337 7718 84081234012344",
            "CONECT1233812339123401234112345",
        )
        results = (
            (12332, [12331]),
            (12333, [12331, 12334]),
            (12334, [12333, 12335]),
            (12335, [12334, 12336]),
            (12336, [12327, 12330, 12335]),
            (12337, [7718, 8408, 12340, 12344]),
            (12338, [12339, 12340, 12341, 12345]),
        )
        for line, res in zip(lines, results):
            bonds = mda.topology.PDBParser._parse_conect(line)
            assert_equal(bonds[0], res[0])
            for bond, res_bond in zip(bonds[1], res[1]):
                assert_equal(bond, res_bond)

    def test_conect_parser_runtime(self):
        with pytest.raises(RuntimeError):
            mda.topology.PDBParser._parse_conect(
                "CONECT12337 7718 " "84081234012344123"
            )

    def test_conect_topo_parser(self):
        """Check that the parser works as intended,
        and that the returned value is a dictionary
        """
        with _PDBPARSER(PDB_conect) as p:
            top = p.parse()
            assert isinstance(top, mda.core.topology.Topology)


def test_conect2ter():
    def parse():
        with PDBParser(PDB_conect2TER) as p:
            struc = p.parse()
        return struc

    with pytest.warns(UserWarning):
        struc = parse()

    assert hasattr(struc, "bonds")
    assert len(struc.bonds.values) == 4


def test_single_conect():
    def parse():
        with PDBParser(PDB_singleconect) as p:
            struc = p.parse()
        return struc

    with pytest.warns(UserWarning):
        struc = parse()
    assert hasattr(struc, "bonds")
    assert len(struc.bonds.values) == 2


def test_new_chainid_new_res():
    # parser must start new residue when chainid starts

    u = mda.Universe(PDB_chainidnewres)

    assert len(u.residues) == 4
    assert_equal(u.residues.resids, [1, 2, 3, 3])
    assert len(u.segments) == 4
    assert_equal(u.segments.segids, ["A", "B", "C", "D"])
    assert len(u.segments[0].atoms) == 5
    assert len(u.segments[1].atoms) == 5
    assert len(u.segments[2].atoms) == 5
    assert len(u.segments[3].atoms) == 7


def test_sameresid_diffresname():
    with _PDBPARSER(PDB_sameresid_diffresname) as p:
        top = p.parse()
    resids = [9, 9]
    resnames = ["GLN", "POPC"]
    for i, (resid, resname) in enumerate(zip(resids, resnames)):
        assert top.resids.values[i] == resid
        assert top.resnames.values[i] == resname


def test_PDB_record_types():
    u = mda.Universe(PDB_HOLE)

    assert u.atoms[0].record_type == "ATOM"
    assert u.atoms[132].record_type == "HETATM"

    assert_equal(u.atoms[10:20].record_types, "ATOM")
    assert_equal(u.atoms[271:].record_types, "HETATM")


PDB_noresid = """\
REMARK For testing reading of CRYST
REMARK This has MODELs then CRYST entries
CRYST1   80.000   80.017   80.017  90.00  90.00  90.00 P 1           1
MODEL        1
ATOM      1  H2  TIP3           10.000  44.891  14.267  1.00  0.00      TIP3
ATOM      2  OH2 TIP3           67.275  48.893  23.568  1.00  0.00      TIP3
ATOM      3  H1  TIP3           66.641  48.181  23.485  1.00  0.00      TIP3
ATOM      4  H2  TIP3           66.986  49.547  22.931  1.00  0.00      TIP3
ENDMDL
"""


def test_PDB_no_resid():
    u = mda.Universe(StringIO(PDB_noresid), format="PDB")

    assert len(u.atoms) == 4
    assert len(u.residues) == 1
    # should have default resid of 1
    assert u.residues[0].resid == 1


PDB_hex = """\
REMARK For testing reading of hex atom numbers
REMARK This has MODELs then hex atom numbers entries
CRYST1   80.000   80.017   80.017  90.00  90.00  90.00 P 1           1
MODEL        1
HETATM    1  H     2 L 400      20.168  00.034  40.428
HETATMA0000  H     2 L 400      40.168  50.034  40.428
HETATMA0001  H     2 L 400      30.453  60.495  50.132
HETATMA0002  H     2 L 400      20.576  40.354  60.483
HETATMA0003  H     2 L 400      10.208  30.067  70.045
END
"""

# this causes an error on Win64/Python 3.8 on Azure when loaded
# in as a file instead
PDB_metals = """\
HETATM    1 CU    CU A   1      00.000  00.000  00.000  1.00 00.00          Cu
HETATM    2 FE    FE A   2      03.000  03.000  03.000  1.00 00.00          Fe
HETATM    3 Ca    Ca A   3      03.000  03.000  03.000  1.00 00.00          Ca
HETATM    3 Mg    Mg A   3      03.000  03.000  03.000  1.00 00.00          Mg
"""


def test_PDB_hex():
    u = mda.Universe(StringIO(PDB_hex), format="PDB")
    assert len(u.atoms) == 5
    assert u.atoms[0].id == 1
    assert u.atoms[1].id == 100000
    assert u.atoms[2].id == 100001
    assert u.atoms[3].id == 100002
    assert u.atoms[4].id == 100003


@pytest.mark.filterwarnings("error:Failed to guess the mass")
def test_PDB_metals():
    u = mda.Universe(StringIO(PDB_metals), format="PDB")

    assert len(u.atoms) == 4
    assert u.atoms[0].mass == pytest.approx(tables.masses["CU"])
    assert u.atoms[1].mass == pytest.approx(tables.masses["FE"])
    assert u.atoms[2].mass == pytest.approx(tables.masses["CA"])
    assert u.atoms[3].mass == pytest.approx(tables.masses["MG"])


def test_PDB_elements():
    """The test checks whether elements attribute are assigned
    properly given a PDB file with valid elements record.
    """
    u = mda.Universe(PDB_elements, format="PDB")
    # fmt: off
    element_list = np.array(
        [
            'N', 'C', 'C', 'O', 'C', 'C', 'O', 'N', 'H', 'H', 'H', 'H', 'H',
            'H', 'H', 'H', 'Cu', 'Fe', 'Mg', 'Ca', 'S', 'O', 'C', 'C', 'S',
            'O', 'C', 'C'
        ],
        dtype=object
    )
    # fmt: on
    assert_equal(u.atoms.elements, element_list)


def test_missing_elements_noattribute():
    """Check that:

    1) a warning is raised if elements are missing
    2) the elements attribute is not set
    """
    wmsg = (
        "Element information is missing, elements attribute will not be "
        "populated"
    )
    with pytest.warns(UserWarning, match=wmsg):
        u = mda.Universe(PDB_small)
    with pytest.raises(AttributeError):
        _ = u.atoms.elements


PDB_wrong_ele = """\
REMARK For testing warnings of wrong elements
REMARK This file represent invalid elements in the elements column
ATOM      1  N   ASN A   1      -8.901   4.127  -0.555  1.00  0.00           N
ATOM      2  CA  ASN A   1      -8.608   3.135  -1.618  1.00  0.00
ATOM      3  C   ASN A   1      -7.117   2.964  -1.897  1.00  0.00           C
ATOM      4  O   ASN A   1      -6.634   1.849  -1.758  1.00  0.00           O
ATOM      5  X   ASN A   1      -9.437   3.396  -2.889  1.00  0.00          XX
TER       6
HETATM    7 CU    CU A   2      03.000  00.000  00.000  1.00 00.00          CU
HETATM    8 FE    FE A   3      00.000  03.000  00.000  1.00 00.00          Fe
HETATM    9 Mg    Mg A   4      03.000  03.000  03.000  1.00 00.00          MG
TER       10
"""


def test_wrong_elements_warnings():
    """The test checks whether there are invalid elements in the elements
    column which have been parsed and returns an appropriate warning.
    """
    with pytest.warns(UserWarning, match="Unknown element XX found"):
        u = mda.Universe(
            StringIO(
                PDB_wrong_ele,
            ),
            format="PDB",
        )

    expected_elements = np.array(
        ["N", "", "C", "O", "", "Cu", "Fe", "Mg"], dtype=object
    )
    gussed_types = np.array(["N", "", "C", "O", "XX", "CU", "Fe", "MG"])
    guseed_masses = np.array(
        [14.007, 0.0, 12.011, 15.999, 0.0, 63.546, 55.847, 24.305], dtype=float
    )

    assert_equal(u.atoms.elements, expected_elements)
    assert_equal(u.atoms.types, gussed_types)
    assert_allclose(u.atoms.masses, guseed_masses)


def test_guessed_masses_and_types_values():
    """Test that guessed masses and types have the expected values for universe
    constructed from PDB file.
    """
    u = mda.Universe(PDB, format="PDB")
    gussed_types = np.array(["N", "H", "H", "H", "C", "H", "C", "H", "H", "C"])
    guseed_masses = [
        14.007,
        1.008,
        1.008,
        1.008,
        12.011,
        1.008,
        12.011,
        1.008,
        1.008,
        12.011,
    ]
    failed_type_guesses = u.atoms.types == ""

    assert_allclose(u.atoms.masses[:10], guseed_masses)
    assert_equal(u.atoms.types[:10], gussed_types)
    assert not failed_type_guesses.any()


def test_nobonds_error():
    """Issue #2832: PDB without CONECT record should not have a bonds
    attribute and raises NoDataError on access"""
    u = mda.Universe(PDB_helix)

    errmsg = "This Universe does not contain bonds information"

    with pytest.raises(NoDataError, match=errmsg):
        u.atoms.bonds


def test_PDB_charges():
    """The test checks whether formalcharges attribute are assigned
    properly given a PDB file with a valid formal charges record.
    """
    u = mda.Universe(PDB_charges)
    # fmt: off
    formal_charges = np.array(
        [
            0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
        ],
        dtype=int
    )
    # fmt: on
    assert_equal(u.atoms.formalcharges, formal_charges)


PDB_charges_nosign = """\
REMARK Invalid charge assignment - no sign for MG2+
HETATM    1 CU    CU A   1      03.000  00.000  00.000  1.00 00.00          CU2+
HETATM    2 FE    FE A   2      00.000  03.000  00.000  1.00 00.00          Fe2+
HETATM    3 Mg    Mg A   3      03.000  03.000  03.000  1.00 00.00          MG2
END
"""


PDB_charges_invertsign = """\
REMARK Invalid charge format for MG2+
HETATM    1 CU    CU A   1      03.000  00.000  00.000  1.00 00.00          CU2+
HETATM    2 FE    FE A   2      00.000  03.000  00.000  1.00 00.00          Fe2+
HETATM    3 Mg    Mg A   3      03.000  03.000  03.000  1.00 00.00          MG+2
END
"""


@pytest.mark.parametrize(
    "infile,entry",
    [[PDB_charges_nosign, r"2"], [PDB_charges_invertsign, r"\+2"]],
)
def test_PDB_bad_charges(infile, entry):
    """
    Test that checks that a warning is raised and formal charges are not set:
        * If there are missing signs for a given formal charge entry
        * If there is an unrecognised formal charge entry
    """
    wmsg = f"Unknown entry {entry} encountered in formal charge field."
    with pytest.warns(UserWarning, match=wmsg):
        u = mda.Universe(StringIO(infile), format="PDB")
        assert not hasattr(u, "formalcharges")


def test_force_chainids_to_segids():
    # get examples from issue 2874 with minor modifications

    res1_str = """\
ATOM    659  N   THR B 315      22.716  15.055  -1.000  1.00 16.08         A N
ATOM    660  CA  THR B 315      22.888  13.803  -0.302  1.00  0.00           C
ATOM    661  C   THR B 315      22.006  12.700  -0.882  1.00  0.00           C
ATOM    662  O   THR B 315      21.138  12.959  -1.727  1.00 16.25         A O
ATOM    663  CB  THR B 315      22.481  13.956   1.182  1.00  0.00           C
ATOM    664  CG2 THR B 315      23.384  14.924   1.927  1.00  0.00           C
ATOM    665  OG1 THR B 315      21.172  14.548   1.274  1.00  0.00           O
"""

    res2_str = """\
ATOM    659  N   THR B 315      22.716  15.055  -1.000  1.00 16.08         A N
ATOM    660  CA  THR B 315      22.888  13.803  -0.302  1.00 15.13         A C
ATOM    661  C   THR B 315      22.006  12.700  -0.882  1.00 15.69         A C
ATOM    662  O   THR B 315      21.138  12.959  -1.727  1.00 16.25         A O
ATOM    663  CB  THR B 315      22.481  13.956   1.182  1.00 16.22         A C
ATOM    664  CG2 THR B 315      22.874  15.310   1.747  1.00 17.32         A C
ATOM    665  OG1 THR B 315      21.047  13.922   1.304  1.00 15.14         A O
"""
    res3_str = """\
ATOM    659  N   THR   315      22.716  15.055  -1.000  1.00 16.08         A N
ATOM    660  CA  THR   315      22.888  13.803  -0.302  1.00 15.13         A C
ATOM    661  C   THR   315      22.006  12.700  -0.882  1.00 15.69         A C
ATOM    662  O   THR   315      21.138  12.959  -1.727  1.00 16.25         A O
ATOM    663  CB  THR   315      22.481  13.956   1.182  1.00 16.22         A C
ATOM    664  CG2 THR   315      22.874  15.310   1.747  1.00 17.32         A C
ATOM    665  OG1 THR   315      21.047  13.922   1.304  1.00 15.14         A O
"""

    res1_not_force = mda.Universe(
        StringIO(res1_str), format="PDB", force_chainids_to_segids=False
    )
    res1_force = mda.Universe(
        StringIO(res1_str), format="PDB", force_chainids_to_segids=True
    )
    res2_not_force = mda.Universe(
        StringIO(res2_str), format="PDB", force_chainids_to_segids=False
    )
    res2_force = mda.Universe(
        StringIO(res2_str), format="PDB", force_chainids_to_segids=True
    )
    res3_force = mda.Universe(
        StringIO(res3_str), format="PDB", force_chainids_to_segids=True
    )

    assert len(res1_not_force.segments) == 4
    assert_equal(["A", "", "A", ""], res1_not_force.segments.segids)
    assert len(res1_not_force.residues) == 4
    assert_equal([315, 315, 315, 315], res1_not_force.residues.resids)
    assert len(res1_force.segments) == 1
    assert_equal(["B"], res1_force.segments.segids)
    assert len(res1_force.residues) == 1
    assert_equal([315], res1_force.residues.resids)
    assert len(res2_not_force.segments) == 1
    assert_equal(["A"], res2_not_force.segments.segids)
    assert len(res2_force.segments) == 1
    assert_equal(["B"], res2_force.segments.segids)
    assert len(res3_force.segments) == 1
    assert_equal([""], res3_force.segments.segids)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
#

import cython
import numpy as np
cimport numpy as cnp
from libc.math cimport sqrt, fabs

from MDAnalysis import NoDataError

from libcpp.set cimport set as cset
from libcpp.map cimport map as cmap
from libcpp.vector cimport vector
from libcpp.utility cimport pair
from cython.operator cimport dereference as deref

cnp.import_array()

__all__ = ['unique_int_1d', 'make_whole', 'find_fragments',
           '_sarrus_det_single', '_sarrus_det_multiple']

cdef extern from "calc_distances.h":
    ctypedef float coordinate[3]
    void minimum_image(double* x, float* box, float* inverse_box)
    void minimum_image_triclinic(double* dx, float* box)

ctypedef cset[int] intset
ctypedef cmap[int, intset] intmap


@cython.boundscheck(False)  # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
def unique_int_1d(cnp.intp_t[:] values):
    """Find the unique elements of a 1D array of integers.

    This function is optimal on sorted arrays.

    Parameters
    ----------
    values: numpy.ndarray
        1D array of dtype ``numpy.int64`` in which to find the unique values.

    Returns
    -------
    numpy.ndarray
        A deduplicated copy of `values`.


    .. versionadded:: 0.19.0
    """
    cdef bint is_monotonic = True
    cdef int i = 0
    cdef int j = 0
    cdef int n_values = values.shape[0]
    cdef cnp.intp_t[:] result = np.empty(n_values, dtype=np.intp)

    if n_values == 0:
        return np.array(result)

    result[0] = values[0]
    for i in range(1, n_values):
        if values[i] != result[j]:
            j += 1
            result[j] = values[i]
        if values[i] < values[i - 1]:
            is_monotonic = False
    result = result[:j + 1]
    if not is_monotonic:
        result = unique_int_1d(np.sort(result))

    return np.array(result)


@cython.boundscheck(False)
def _in2d(cnp.intp_t[:, :] arr1, cnp.intp_t[:, :] arr2):
    """Similar to np.in1d except works on 2d arrays

    Parameters
    ----------
    arr1, arr2 : numpy.ndarray, shape (n,2) and (m, 2)
       arrays of integers

    Returns
    -------
    in1d : bool array
      if an element of arr1 was in arr2

    .. versionadded:: 1.1.0
    """
    cdef object out
    cdef ssize_t i
    cdef cset[pair[cnp.intp_t, cnp.intp_t]] hits
    cdef pair[cnp.intp_t, cnp.intp_t] p

    """
    Construct a set from arr2 called hits
    then for each entry in arr1, check if there's a hit in this set

    python would look like:

    hits = {(i, j) for (i, j) in arr2}
    results = np.empty(arr1.shape[0])
    for i, (x, y) in enumerate(arr1):
        results[i] = (x, y) in hits

    return results
    """
    if not arr1.shape[1] == 2 or not arr2.shape[1] == 2:
        raise ValueError("Both arrays must be (n, 2) arrays")

    for i in range(arr2.shape[0]):
        p = pair[cnp.intp_t, cnp.intp_t](arr2[i, 0], arr2[i, 1])
        hits.insert(p)

    out = np.empty(arr1.shape[0], dtype=np.uint8)
    cdef unsigned char[::1] results = out
    for i in range(arr1.shape[0]):
        p = pair[cnp.intp_t, cnp.intp_t](arr1[i, 0], arr1[i, 1])

        if hits.count(p):
            results[i] = True
        else:
            results[i] = False

    return out.astype(bool)


cdef intset difference(intset a, intset b):
    """a.difference(b)

    Returns set of values in a which are not in b
    """
    cdef intset output
    for val in a:
        if b.count(val) != 1:
            output.insert(val)
    return output


@cython.boundscheck(False)
@cython.wraparound(False)
def make_whole(atomgroup, reference_atom=None, inplace=True):
    """Move all atoms in a single molecule so that bonds don't split over
    images.

    This function is most useful when atoms have been packed into the primary
    unit cell, causing breaks mid molecule, with the molecule then appearing
    on either side of the unit cell. This is problematic for operations
    such as calculating the center of mass of the molecule. ::

       +-----------+     +-----------+
       |           |     |           |
       | 6       3 |     |         3 | 6
       | !       ! |     |         ! | !
       |-5-8   1-2-| ->  |       1-2-|-5-8
       | !       ! |     |         ! | !
       | 7       4 |     |         4 | 7
       |           |     |           |
       +-----------+     +-----------+


    Parameters
    ----------
    atomgroup : AtomGroup
        The :class:`MDAnalysis.core.groups.AtomGroup` to work with.
        The positions of this are modified in place.  All these atoms
        must belong to the same molecule or fragment.
    reference_atom : :class:`~MDAnalysis.core.groups.Atom`
        The atom around which all other atoms will be moved.
        Defaults to atom 0 in the atomgroup.
    inplace : bool, optional
        If ``True``, coordinates are modified in place.

    Returns
    -------
    coords : numpy.ndarray
        The unwrapped atom coordinates.

    Raises
    ------
    NoDataError
        There are no bonds present.
        (See :func:`~MDAnalysis.topology.core.guess_bonds`)

    ValueError
        The algorithm fails to work.  This is usually
        caused by the atomgroup not being a single fragment.
        (ie the molecule can't be traversed by following bonds)


    Example
    -------
    Make fragments whole::

        from MDAnalysis.lib.mdamath import make_whole

        # This algorithm requires bonds, these can be guessed!
        u = mda.Universe(......, guess_bonds=True)

        # MDAnalysis can split AtomGroups into their fragments
        # based on bonding information.
        # Note that this function will only handle a single fragment
        # at a time, necessitating a loop.
        for frag in u.atoms.fragments:
            make_whole(frag)

    Alternatively, to keep a single atom in place as the anchor::

        # This will mean that atomgroup[10] will NOT get moved,
        # and all other atoms will move (if necessary).
        make_whole(atomgroup, reference_atom=atomgroup[10])


    See Also
    --------
    :meth:`MDAnalysis.core.groups.AtomGroup.unwrap`


    .. versionadded:: 0.11.0
    .. versionchanged:: 0.20.0
        Inplace-modification of atom positions is now optional, and positions
        are returned as a numpy array.
    """
    cdef intset refpoints, todo, done
    cdef cnp.intp_t i, j, nloops, ref, atom, other, natoms
    cdef cmap[int, int] ix_to_rel
    cdef intmap bonding
    cdef int[:, :] bonds
    cdef float[:, :] oldpos, newpos
    cdef bint ortho
    cdef float[:] box
    cdef float[:, :] tri_box
    cdef float half_box[3]
    cdef float inverse_box[3]
    cdef double vec[3]
    cdef ssize_t[:] ix_view
    cdef bint is_unwrapped

    # map of global indices to local indices
    ix_view = atomgroup.ix[:]
    natoms = atomgroup.ix.shape[0]

    oldpos = atomgroup.positions

    # Nothing to do for less than 2 atoms
    if natoms < 2:
        return np.array(oldpos)

    for i in range(natoms):
        ix_to_rel[ix_view[i]] = i

    if reference_atom is None:
        ref = 0
    else:
        # Sanity check
        if reference_atom not in atomgroup:
            raise ValueError("Reference atom not in atomgroup")
        ref = ix_to_rel[reference_atom.ix]

    if atomgroup.dimensions is None:
        raise ValueError("No box information available. "
                         "You can set dimensions using 'atomgroup.dimensions='")
    box = atomgroup.dimensions
    for i in range(3):
        half_box[i] = 0.5 * box[i]

    ortho = True
    for i in range(3, 6):
        if box[i] != 90.0:
            ortho = False

    if ortho:
        # If atomgroup is already unwrapped, bail out
        is_unwrapped = True
        for i in range(1, natoms):
            for j in range(3):
                if fabs(oldpos[i, j] - oldpos[0, j]) >= half_box[j]:
                    is_unwrapped = False
                    break
            if not is_unwrapped:
                break
        if is_unwrapped:
            return np.array(oldpos)
        for i in range(3):
            inverse_box[i] = 1.0 / box[i]
    else:
        from .mdamath import triclinic_vectors
        tri_box = triclinic_vectors(box)

    # C++ dict of bonds
    try:
        bonds = atomgroup.bonds.to_indices()
    except (AttributeError, NoDataError):
        raise NoDataError("The atomgroup is required to have bonds")
    for i in range(bonds.shape[0]):
        atom = bonds[i, 0]
        other = bonds[i, 1]
        # only add bonds if both atoms are in atoms set
        if ix_to_rel.count(atom) and ix_to_rel.count(other):
            atom = ix_to_rel[atom]
            other = ix_to_rel[other]

            bonding[atom].insert(other)
            bonding[other].insert(atom)

    newpos = np.zeros((oldpos.shape[0], 3), dtype=np.float32)

    refpoints = intset()  # Who is safe to use as reference point?
    done = intset()  # Who have I already searched around?
    # initially we have one starting atom whose position is in correct image
    refpoints.insert(ref)
    for i in range(3):
        newpos[ref, i] = oldpos[ref, i]

    nloops = 0
    while <cnp.intp_t> refpoints.size() < natoms and nloops < natoms:
        # count iterations to prevent infinite loop here
        nloops += 1

        # We want to iterate over atoms that are good to use as reference
        # points, but haven't been searched yet.
        todo = difference(refpoints, done)
        for atom in todo:
            for other in bonding[atom]:
                # If other is already a refpoint, leave alone
                if refpoints.count(other):
                    continue
                # Draw vector from atom to other
                for i in range(3):
                    vec[i] = oldpos[other, i] - newpos[atom, i]
                # Apply periodic boundary conditions to this vector
                if ortho:
                    minimum_image(&vec[0], &box[0], &inverse_box[0])
                else:
                    minimum_image_triclinic(&vec[0], &tri_box[0, 0])
                # Then define position of other based on this vector
                for i in range(3):
                    newpos[other, i] = newpos[atom, i] + vec[i]

                # This other atom can now be used as a reference point
                refpoints.insert(other)
            done.insert(atom)

    if <cnp.intp_t> refpoints.size() < natoms:
        raise ValueError("AtomGroup was not contiguous from bonds, process failed")
    if inplace:
        atomgroup.positions = newpos
    return np.array(newpos)


@cython.boundscheck(False)
@cython.wraparound(False)
cdef float _dot(float * a, float * b):
    """Return dot product of two 3d vectors"""
    cdef ssize_t n
    cdef float sum1

    sum1 = 0.0
    for n in range(3):
        sum1 += a[n] * b[n]
    return sum1


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void _cross(float * a, float * b, float * result):
    """
    Calculates the cross product between 3d vectors

    Note
    ----
    Modifies the result array
    """

    result[0] = a[1]*b[2] - a[2]*b[1]
    result[1] = - a[0]*b[2] + a[2]*b[0]
    result[2] = a[0]*b[1] - a[1]*b[0]

cdef float _norm(float * a):
    """
    Calculates the magnitude of the vector
    """
    cdef float result
    cdef ssize_t n
    result = 0.0
    for n in range(3):
        result += a[n]*a[n]
    return sqrt(result)


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef cnp.float64_t _sarrus_det_single(cnp.float64_t[:, ::1] m):
    """Computes the determinant of a 3x3 matrix."""
    cdef cnp.float64_t det
    det = m[0, 0] * m[1, 1] * m[2, 2]
    det -= m[0, 0] * m[1, 2] * m[2, 1]
    det += m[0, 1] * m[1, 2] * m[2, 0]
    det -= m[0, 1] * m[1, 0] * m[2, 2]
    det += m[0, 2] * m[1, 0] * m[2, 1]
    det -= m[0, 2] * m[1, 1] * m[2, 0]
    return det


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef cnp.ndarray _sarrus_det_multiple(cnp.float64_t[:, :, ::1] m):
    """Computes all determinants of an array of 3x3 matrices."""
    cdef cnp.intp_t n
    cdef cnp.intp_t i
    cdef cnp.float64_t[:] det
    n = m.shape[0]
    det = np.empty(n, dtype=np.float64)
    for i in range(n):
        det[i] = m[i, 0, 0] * m[i, 1, 1] * m[i, 2, 2]
        det[i] -= m[i, 0, 0] * m[i, 1, 2] * m[i, 2, 1]
        det[i] += m[i, 0, 1] * m[i, 1, 2] * m[i, 2, 0]
        det[i] -= m[i, 0, 1] * m[i, 1, 0] * m[i, 2, 2]
        det[i] += m[i, 0, 2] * m[i, 1, 0] * m[i, 2, 1]
        det[i] -= m[i, 0, 2] * m[i, 1, 1] * m[i, 2, 0]
    return np.array(det)


@cython.boundscheck(False)
@cython.wraparound(False)
def find_fragments(atoms, bondlist):
    """Calculate distinct fragments from nodes (atom indices) and edges (pairs
    of atom indices).

    Parameters
    ----------
    atoms : array_like
       1-D Array of atom indices (dtype will be converted to ``numpy.int64``
       internally)
    bonds : array_like
       2-D array of bonds (dtype will be converted to ``numpy.int32``
       internally), where ``bonds[i, 0]`` and ``bonds[i, 1]`` are the
       indices of atoms connected by the ``i``-th bond. Any bonds referring to
       atom indices not in `atoms` will be ignored.

    Returns
    -------
    fragments : list
       List of arrays, each containing the atom indices of a fragment.

    .. versionaddded:: 0.19.0
    """
    cdef intmap bondmap
    cdef intset todo, frag_todo, frag_done
    cdef vector[int] this_frag
    cdef int i, a, b
    cdef cnp.int64_t[:] atoms_view
    cdef cnp.int32_t[:, :] bonds_view

    atoms_view = np.asarray(atoms, dtype=np.int64)
    bonds_view = np.asarray(bondlist, dtype=np.int32)

    # grab record of which atoms I have to process
    # ie set of all nodes
    for i in range(atoms_view.shape[0]):
        todo.insert(atoms_view[i])
    # Process edges into map
    for i in range(bonds_view.shape[0]):
        a = bonds_view[i, 0]
        b = bonds_view[i, 1]
        # only include edges if both are known nodes
        if todo.count(a) and todo.count(b):
            bondmap[a].insert(b)
            bondmap[b].insert(a)

    frags = []

    while not todo.empty():  # While not all nodes have been done
        # Start a new fragment
        frag_todo.clear()
        frag_done.clear()
        this_frag.clear()
        # Grab a start point for next fragment
        frag_todo.insert(deref(todo.begin()))

        # Loop until fragment fully explored
        while not frag_todo.empty():
            # Pop next in this frag todo
            a = deref(frag_todo.begin())
            frag_todo.erase(a)
            if not frag_done.count(a):
                this_frag.push_back(a)
                frag_done.insert(a)
                todo.erase(a)
                for b in bondmap[a]:
                    if not frag_done.count(b):
                        frag_todo.insert(b)

        # Add fragment to output
        frags.append(np.asarray(this_frag))

    return frags



import MDAnalysis

try:
    from MDAnalysisTests.datafiles import TPR, XTC
except:
    pass

try:
    from MDAnalysis.analysis.rdf import InterRDF, InterRDF_s
except:
    pass


class SimpleRdfBench(object):
    """Benchmarks for MDAnalysis.analysis.rdf.InterRDF"""

    params = ([20, 75, 200], [[0, 5], [0, 15], [0, 20]], [1, 100, 1000, 10000])

    param_names = ["nbins", "range_val", "natoms"]

    def setup(self, nbins, range_val, natoms):

        self.sel_str = "name OW"

        self.u = MDAnalysis.Universe(TPR, XTC)

        try:
            self.sel = self.u.select_atoms(self.sel_str)[:natoms]
        except AttributeError:
            self.sel = self.u.selectAtoms(self.sel_str)[:natoms]

        # do not include initialization of the
        # InterRDF object in the benchmark itself

        self.rdf = InterRDF(g1=self.sel, g2=self.sel, nbins=nbins, range=range_val)

    def time_interrdf(self, nbins, range_val, natoms):
        """Benchmark a full trajectory parse
        by MDAnalysis.analysis.rdf.InterRDF
        """
        self.rdf.run()

class SimpleRdfsBench(object):
    """Benchmarks for MDAnalysis.analysis.rdf.InterRDF_s"""

    params = ([20, 75, 200], [[0, 5], [0, 15], [0, 20]], [10, 100], [1, 3, 9])

    param_names = ["nbins", "range_val", "natoms", "npairs"]

    def setup(self, nbins, range_val, natoms, npairs):

        self.sel_str = "name OW"

        self.u = MDAnalysis.Universe(TPR, XTC)

        try:
            self.sel = self.u.select_atoms(self.sel_str)[:natoms]
        except AttributeError:
            self.sel = self.u.selectAtoms(self.sel_str)[:natoms]

        ags = [[self.sel, self.sel]] * npairs
        self.rdf_s = InterRDF_s(self.u, ags, nbins=nbins, range=range_val)


    def time_interrdfs(self, nbins, range_val, natoms, npairs):
        """Benchmark a full trajectory parse
        by MDAnalysis.analysis.rdf.InterRDF_s
        """
        self.rdf_s.run()


# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
import sys
from numpy.testing import assert_equal


def test_failed_import(monkeypatch):
    # Putting this test first to avoid datafiles already being loaded
    errmsg = "MDAnalysisTests package not installed."

    monkeypatch.setitem(sys.modules, "MDAnalysisTests.datafiles", None)

    if "MDAnalysis.tests.datafiles" in sys.modules:
        monkeypatch.delitem(sys.modules, "MDAnalysis.tests.datafiles")

    with pytest.raises(ImportError, match=errmsg):
        import MDAnalysis.tests.datafiles


def test_import():
    try:
        import MDAnalysis.tests.datafiles
    except ImportError:
        pytest.fail(
            "Failed to 'import MDAnalysis.tests.datafiles --- install MDAnalysisTests"
        )


def test_all_exports():
    import MDAnalysisTests.datafiles

    missing = [
        name
        for name in dir(MDAnalysisTests.datafiles)
        if not name.startswith("_")
        and name not in MDAnalysisTests.datafiles.__all__
        and name != "MDAnalysisTests"
    ]
    assert_equal(missing, [], err_msg="Variables need to be added to __all__.")


def test_export_variables():
    import MDAnalysisTests.datafiles
    import MDAnalysis.tests.datafiles

    missing = [
        name
        for name in MDAnalysisTests.datafiles.__all__
        if name not in dir(MDAnalysis.tests.datafiles)
    ]
    assert_equal(
        missing,
        [],
        err_msg="Variables not exported to MDAnalysis.tests.datafiles",
    )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
PDBQT topology parser
=====================

Use a PDBQT_ file to build a minimum internal structure representation (list of
atoms), including AutoDock_ atom types (stored as :attr:`Atom.type`) and
partial charges (:attr:`Atom.charge`).

* Reads a PDBQT file line by line and does not require sequential atom numbering.
* Multi-model PDBQT files are not supported.

.. note::

    By default, masses will be guessed on Universe creation.
    This may change in release 3.0.
    See :ref:`Guessers` for more information.

Notes
-----
Only reads atoms and their names; connectivity is not
deduced.


See Also
--------
:class:`MDAnalysis.coordinates.PDBQT`


Classes
-------
.. autoclass:: PDBQTParser
   :members:
   :inherited-members:


.. _PDBQT:
   https://autodock.scripps.edu/wp-content/uploads/sites/56/2021/10/AutoDock4.2.6_UserGuide.pdf 
.. _AutoDock:
   http://autodock.scripps.edu/
"""
import numpy as np

from ..lib import util
from .base import TopologyReaderBase, change_squash
from ..core.topology import Topology
from ..core.topologyattrs import (
    Atomids,
    Atomnames,
    AltLocs,
    Atomtypes,
    Charges,
    ICodes,
    Occupancies,
    RecordTypes,
    Resids,
    Resnums,
    Resnames,
    Segids,
    ChainIDs,
    Tempfactors,
)


class PDBQTParser(TopologyReaderBase):
    """Read topology from a PDBQT file.

    Creates the following Attributes:
     - atom ids (serial)
     - atom types
     - atom names
     - altLocs
     - resnames
     - chainIDs (assigned to segid as well)
     - resids
     - record_types (ATOM/HETATM)
     - icodes
     - occupancies
     - tempfactors
     - charges


    .. versionchanged:: 0.18.0
       Added parsing of Record types
    .. versionchanged:: 2.7.0
       Columns 67 - 70 in ATOM records, corresponding to the field *footnote*,
       are now ignored. See Autodock's `reference`_.
    .. versionchanged:: 2.8.0
        Removed mass guessing (attributes guessing takes place now
        through universe.guess_TopologyAttrs() API).

       .. _reference:
          https://autodock.scripps.edu/wp-content/uploads/sites/56/2021/10/AutoDock4.2.6_UserGuide.pdf
    """

    format = "PDBQT"

    def parse(self, **kwargs):
        """Parse atom information from PDBQT file *filename*.

        Returns
        -------
        MDAnalysis Topology object
        """
        record_types = []
        serials = []
        names = []
        altlocs = []
        resnames = []
        chainids = []
        resids = []
        icodes = []
        occupancies = []
        tempfactors = []
        charges = []
        atomtypes = []

        with util.openany(self.filename) as f:
            for line in f:
                line = line.strip()
                if not line.startswith(("ATOM", "HETATM")):
                    continue
                record_types.append(line[:6].strip())
                serials.append(int(line[6:11]))
                names.append(line[12:16].strip())
                altlocs.append(line[16:17].strip())
                resnames.append(line[17:21].strip())
                chainids.append(line[21:22].strip())
                resids.append(int(line[22:26]))
                icodes.append(line[26:27].strip())
                occupancies.append(float(line[54:60]))
                tempfactors.append(float(line[60:66]))
                charges.append(float(line[70:76]))
                atomtypes.append(line[77:80].strip())

        n_atoms = len(serials)

        attrs = []
        for attrlist, Attr, dtype in (
            (record_types, RecordTypes, object),
            (serials, Atomids, np.int32),
            (names, Atomnames, object),
            (altlocs, AltLocs, object),
            (occupancies, Occupancies, np.float32),
            (tempfactors, Tempfactors, np.float32),
            (charges, Charges, np.float32),
            (atomtypes, Atomtypes, object),
        ):
            attrs.append(Attr(np.array(attrlist, dtype=dtype)))

        resids = np.array(resids, dtype=np.int32)
        icodes = np.array(icodes, dtype=object)
        resnames = np.array(resnames, dtype=object)
        chainids = np.array(chainids, dtype=object)

        attrs.append(ChainIDs(chainids))

        residx, (resids, icodes, resnames, chainids) = change_squash(
            (resids, icodes), (resids, icodes, resnames, chainids)
        )
        n_residues = len(resids)
        attrs.append(Resids(resids))
        attrs.append(Resnums(resids.copy()))
        attrs.append(ICodes(icodes))
        attrs.append(Resnames(resnames))

        segidx, (segids,) = change_squash((chainids,), (chainids,))
        n_segments = len(segids)
        attrs.append(Segids(segids))

        top = Topology(
            n_atoms,
            n_residues,
            n_segments,
            attrs=attrs,
            atom_resindex=residx,
            residue_segindex=segidx,
        )

        return top



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest

import numpy as np
from numpy.testing import assert_equal, assert_almost_equal

from MDAnalysis import units


class TestUnitEncoding(object):
    def test_unicode(self):
        try:
            assert_equal(units.lengthUnit_factor["\u212b"], 1.0)
        except KeyError:
            raise AssertionError("Unicode symbol for Angtrom not supported")

    def test_unicode_encoding_with_symbol(self):
        try:
            assert_equal(units.lengthUnit_factor["Å"], 1.0)
        except KeyError:
            raise AssertionError(
                "UTF-8-encoded symbol for Angtrom not supported"
            )


class TestConstants(object):
    # CODATA 2010 (NIST): http://physics.nist.gov/cuu/Constants/
    # (accessed 2015-02-15)
    # Add a reference value to this dict for every entry in
    # units.constants
    constants_reference = (
        ("N_Avogadro", 6.02214129e23),  # mol**-1
        ("elementary_charge", 1.602176565e-19),  # As
        ("calorie", 4.184),  # J
        ("Boltzmann_constant", 8.314462159e-3),  # KJ (mol K)**-1
        ("Boltzman_constant", 8.314462159e-3),  # remove in 2.8.0
        ("electric_constant", 5.526350e-3),  # As (Angstroms Volts)**-1
    )

    @pytest.mark.parametrize("name, value", constants_reference)
    def test_constant(self, name, value):
        assert_almost_equal(units.constants[name], value)

    def test_boltzmann_typo_deprecation(self):
        wmsg = (
            "Please use 'Boltzmann_constant' henceforth. The key "
            "'Boltzman_constant' was a typo and will be removed "
            "in MDAnalysis 2.8.0."
        )
        with pytest.warns(DeprecationWarning, match=wmsg):
            units.constants["Boltzman_constant"]


class TestConversion(object):
    @staticmethod
    def _assert_almost_equal_convert(value, u1, u2, ref):
        val = units.convert(value, u1, u2)
        assert_almost_equal(
            val, ref, err_msg="Conversion {0} --> {1} failed".format(u1, u2)
        )

    nm = 12.34567
    A = nm * 10.0

    @pytest.mark.parametrize(
        "quantity, unit1, unit2, ref",
        (
            (nm, "nm", "A", A),
            (A, "Angstrom", "nm", nm),
            (A, "\u212b", "nm", nm),
        ),
    )
    def test_length(self, quantity, unit1, unit2, ref):
        self._assert_almost_equal_convert(quantity, unit1, unit2, ref)

    @pytest.mark.parametrize(
        "quantity, unit1, unit2, ref",
        (
            (1, "ps", "AKMA", 20.45482949774598),
            (1, "AKMA", "ps", 0.04888821),
            (1, "ps", "ms", 1e-9),
            (1, "ms", "ps", 1e9),
            (1, "ps", "us", 1e-6),
            (1, "us", "ps", 1e6),
            (1, "\u03BCs", "ps", 1e6),
            (1, "fs", "ns", 1e-6),
        ),
    )
    def test_time(self, quantity, unit1, unit2, ref):
        self._assert_almost_equal_convert(quantity, unit1, unit2, ref)

    @pytest.mark.parametrize(
        "quantity, unit1, unit2, ref",
        (
            (1, "kcal/mol", "kJ/mol", 4.184),
            (1, "kcal/mol", "eV", 0.0433641),
        ),
    )
    def test_energy(self, quantity, unit1, unit2, ref):
        self._assert_almost_equal_convert(quantity, unit1, unit2, ref)

    @pytest.mark.parametrize(
        "quantity, unit1, unit2, ref",
        (
            (1, "kJ/(mol*A)", "J/m", 1.66053892103219e-11),
            (2.5, "kJ/(mol*nm)", "kJ/(mol*A)", 0.25),
            (1, "kcal/(mol*Angstrom)", "kJ/(mol*Angstrom)", 4.184),
        ),
    )
    def test_force(self, quantity, unit1, unit2, ref):
        self._assert_almost_equal_convert(quantity, unit1, unit2, ref)

    @pytest.mark.parametrize(
        "quantity, unit1, unit2, ref",
        (
            (1, "A/ps", "m/s", 1e-10 / 1e-12),
            (1, "A/ps", "nm/ps", 0.1),
            (1, "A/ps", "pm/ps", 1e2),
            (1, "A/ms", "A/ps", 1e-9),
            (1, "A/us", "A/ps", 1e-6),
            (1, "A/\u03BCs", "A/ps", 1e-6),
            (1, "\u212b/\u03BCs", "A/ps", 1e-6),
            (1, "A/fs", "A/ps", 1e3),
            (1, "A/ps", "A/fs", 1e-3),
            (1, "A/ns", "A/ms", 1e6),
            (1, "\u212b/ns", "A/ms", 1e6),
            (1, "A/ps", "A/ps", 1.0),
            (1, "nm/ns", "pm/ps", 1.0),
            (1, "m/s", "A/fs", 1e-5),
            (1, "A/AKMA", "A/ps", 1 / 4.888821e-2),
        ),
    )
    def test_speed(self, quantity, unit1, unit2, ref):
        self._assert_almost_equal_convert(quantity, unit1, unit2, ref)

    @pytest.mark.parametrize(
        "quantity, unit1, unit2", ((nm, "Stone", "nm"), (nm, "nm", "Stone"))
    )
    def test_unit_unknown(self, quantity, unit1, unit2):
        with pytest.raises(ValueError):
            units.convert(quantity, unit1, unit2)

    def test_unit_unconvertable(self):
        nm = 12.34567
        A = nm * 10.0
        with pytest.raises(ValueError):
            units.convert(A, "A", "ps")


class TestBaseUnits:
    @staticmethod
    @pytest.fixture
    def ref():
        # This is a copy of the dictionary we expect.
        # We want to know if base units are added or altered.
        ref = {
            "length": "A",
            "time": "ps",
            "energy": "kJ/mol",
            "charge": "e",
            "force": "kJ/(mol*A)",
            "speed": "A/ps",
        }
        return ref

    def test_MDANALYSIS_BASE_UNITS_correct(self, ref):
        assert ref == units.MDANALYSIS_BASE_UNITS



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
import os

import pytest
from numpy.testing import assert_allclose, assert_equal, assert_almost_equal

import MDAnalysis as mda
from MDAnalysisTests.datafiles import NAMDBIN, PDB_small
from MDAnalysisTests.coordinates.base import (
    _SingleFrameReader,
    BaseReference,
    BaseWriterTest,
)


class TestNAMDBINReader(_SingleFrameReader):
    """Test reading namd binary coordinate file"""

    __test__ = True

    def setUp(self):
        # can lead to race conditions when testing in parallel
        self.universe = mda.Universe(PDB_small, NAMDBIN)
        # 6 decimals in NAMDBIN spec
        self.prec = 6
        self.ref_n_atoms = 3341

    def test_parse_n_atoms(self):
        n_atoms = mda.coordinates.NAMDBIN.NAMDBINReader.parse_n_atoms(NAMDBIN)
        assert_equal(n_atoms, self.ref_n_atoms)

    def test_get_writer_from_reader(self):
        universe = mda.Universe(PDB_small, NAMDBIN)
        writer = universe.trajectory.Writer("NAMDBIN-test")
        assert isinstance(writer, mda.coordinates.NAMDBIN.NAMDBINWriter)


class NAMDBINReference(BaseReference):
    def __init__(self):
        super(NAMDBINReference, self).__init__()
        self.trajectory = NAMDBIN
        self.topology = PDB_small
        self.reader = mda.coordinates.NAMDBIN.NAMDBINReader
        self.writer = mda.coordinates.NAMDBIN.NAMDBINWriter
        self.ext = "coor"
        self.volume = 0
        self.dimensions = np.zeros(6)
        self.container_format = True


class NAMDBINWriter(BaseWriterTest):
    __test__ = True

    @staticmethod
    @pytest.fixture()
    def ref():
        return NAMDBINReference()



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
from io import StringIO
import pytest
import os
import warnings
import re
import textwrap
from unittest.mock import Mock, patch
import sys
import copy
import shutil

import numpy as np
from numpy.testing import (
    assert_equal,
    assert_almost_equal,
    assert_array_almost_equal,
    assert_array_equal,
    assert_allclose,
)
from itertools import combinations_with_replacement as comb_wr

import MDAnalysis as mda
import MDAnalysis.lib.util as util
from MDAnalysis.lib.util import (
    cached,
    static_variables,
    warn_if_not_unique,
    check_coords,
    store_init_arguments,
    check_atomgroup_not_empty,
)
from MDAnalysis.core.topologyattrs import Bonds
from MDAnalysis.exceptions import NoDataError, DuplicateWarning
from MDAnalysis.core.groups import AtomGroup
from MDAnalysisTests.datafiles import (
    PSF,
    DCD,
    Make_Whole,
    TPR,
    GRO,
    fullerene,
    two_water_gro,
)


def test_absence_cutil():
    with patch.dict("sys.modules", {"MDAnalysis.lib._cutil": None}):
        import importlib

        with pytest.raises(ImportError):
            importlib.reload(sys.modules["MDAnalysis.lib.util"])


def test_presence_cutil():
    mock = Mock()
    with patch.dict("sys.modules", {"MDAnalysis.lib._cutil": mock}):
        try:
            import MDAnalysis.lib._cutil
        except ImportError:
            pytest.fail(
                msg="""MDAnalysis.lib._cutil should not raise
                         an ImportError if cutil is available."""
            )


def convert_aa_code_long_data():
    aa = [
        (
            "H",
            (
                "HIS",
                "HISA",
                "HISB",
                "HSE",
                "HSD",
                "HIS1",
                "HIS2",
                "HIE",
                "HID",
            ),
        ),
        ("K", ("LYS", "LYSH", "LYN")),
        ("A", ("ALA",)),
        ("D", ("ASP", "ASPH", "ASH")),
        ("E", ("GLU", "GLUH", "GLH")),
        ("N", ("ASN",)),
        ("Q", ("GLN",)),
        ("C", ("CYS", "CYSH", "CYS1", "CYS2")),
    ]
    for resname1, strings in aa:
        for resname3 in strings:
            yield (resname3, resname1)


class TestStringFunctions(object):
    # (1-letter, (canonical 3 letter, other 3/4 letter, ....))
    aa = [
        (
            "H",
            (
                "HIS",
                "HISA",
                "HISB",
                "HSE",
                "HSD",
                "HIS1",
                "HIS2",
                "HIE",
                "HID",
            ),
        ),
        ("K", ("LYS", "LYSH", "LYN")),
        ("A", ("ALA",)),
        ("D", ("ASP", "ASPH", "ASH")),
        ("E", ("GLU", "GLUH", "GLH")),
        ("N", ("ASN",)),
        ("Q", ("GLN",)),
        ("C", ("CYS", "CYSH", "CYS1", "CYS2")),
    ]

    residues = [
        ("LYS300:HZ1", ("LYS", 300, "HZ1")),
        ("K300:HZ1", ("LYS", 300, "HZ1")),
        ("K300", ("LYS", 300, None)),
        ("LYS 300:HZ1", ("LYS", 300, "HZ1")),
        ("M1:CA", ("MET", 1, "CA")),
    ]

    @pytest.mark.parametrize("rstring, residue", residues)
    def test_parse_residue(self, rstring, residue):
        assert util.parse_residue(rstring) == residue

    def test_parse_residue_ValueError(self):
        with pytest.raises(ValueError):
            util.parse_residue("ZZZ")

    @pytest.mark.parametrize("resname3, resname1", convert_aa_code_long_data())
    def test_convert_aa_3to1(self, resname3, resname1):
        assert util.convert_aa_code(resname3) == resname1

    @pytest.mark.parametrize("resname1, strings", aa)
    def test_convert_aa_1to3(self, resname1, strings):
        assert util.convert_aa_code(resname1) == strings[0]

    @pytest.mark.parametrize("x", ("XYZXYZ", "£"))
    def test_ValueError(self, x):
        with pytest.raises(ValueError):
            util.convert_aa_code(x)


def test_greedy_splitext(
    inp="foo/bar/boing.2.pdb.bz2", ref=["foo/bar/boing", ".2.pdb.bz2"]
):
    inp = os.path.normpath(inp)
    ref[0] = os.path.normpath(ref[0])
    ref[1] = os.path.normpath(ref[1])
    root, ext = util.greedy_splitext(inp)
    assert root == ref[0], "root incorrect"
    assert ext == ref[1], "extension incorrect"


@pytest.mark.parametrize(
    "iterable, value",
    [
        ([1, 2, 3], True),
        ([], True),
        ((1, 2, 3), True),
        ((), True),
        (range(3), True),
        (np.array([1, 2, 3]), True),
        (123, False),
        ("byte string", False),
        ("unicode string", False),
    ],
)
def test_iterable(iterable, value):
    assert util.iterable(iterable) == value


class TestFilename(object):
    root = "foo"
    filename = "foo.psf"
    ext = "pdb"
    filename2 = "foo.pdb"

    @pytest.mark.parametrize(
        "name, ext, keep, actual_name",
        [
            (filename, None, False, filename),
            (filename, ext, False, filename2),
            (filename, ext, True, filename),
            (root, ext, False, filename2),
            (root, ext, True, filename2),
        ],
    )
    def test_string(self, name, ext, keep, actual_name):
        file_name = util.filename(name, ext, keep)
        assert file_name == actual_name

    def test_named_stream(self):
        ns = util.NamedStream(StringIO(), self.filename)
        fn = util.filename(ns, ext=self.ext)
        # assert_equal replace by this if loop to avoid segfault on some systems
        if fn != ns:
            pytest.fail("fn and ns are different")
        assert str(fn) == self.filename2
        assert ns.name == self.filename2


class Class_with_Caches(object):
    def __init__(self):
        self._cache = dict()
        self.ref1 = 1.0
        self.ref2 = 2.0
        self.ref3 = 3.0
        self.ref4 = 4.0
        self.ref5 = 5.0
        self.ref6 = 6.0
        # For universe-validated caches
        # One-line lambda-like class
        self.universe = type("Universe", (), dict())()
        self.universe._cache = {"_valid": {}}

    @cached("val1")
    def val1(self):
        return self.ref1

    # Do one with property decorator as these are used together often
    @property
    @cached("val2")
    def val2(self):
        return self.ref2

    # Check use of property setters
    @property
    @cached("val3")
    def val3(self):
        return self.ref3

    @val3.setter
    def val3(self, new):
        self._clear_caches("val3")
        self._fill_cache("val3", new)

    @val3.deleter
    def val3(self):
        self._clear_caches("val3")

    # Check that args are passed through to underlying functions
    @cached("val4")
    def val4(self, n1, n2):
        return self._init_val_4(n1, n2)

    def _init_val_4(self, m1, m2):
        return self.ref4 + m1 + m2

    # Args and Kwargs
    @cached("val5")
    def val5(self, n, s=None):
        return self._init_val_5(n, s=s)

    def _init_val_5(self, n, s=None):
        return n * s

    # Property decorator and universally-validated cache
    @property
    @cached("val6", universe_validation=True)
    def val6(self):
        return self.ref5 + 1.0

    # These are designed to mimic the AG and Universe cache methods
    def _clear_caches(self, *args):
        if len(args) == 0:
            self._cache = dict()
        else:
            for name in args:
                try:
                    del self._cache[name]
                except KeyError:
                    pass

    def _fill_cache(self, name, value):
        self._cache[name] = value


class TestCachedDecorator(object):
    @pytest.fixture()
    def obj(self):
        return Class_with_Caches()

    def test_val1_lookup(self, obj):
        obj._clear_caches()
        assert "val1" not in obj._cache
        assert obj.val1() == obj.ref1
        ret = obj.val1()
        assert "val1" in obj._cache
        assert obj._cache["val1"] == ret
        assert obj.val1() is obj._cache["val1"]

    def test_val1_inject(self, obj):
        # Put something else into the cache and check it gets returned
        # this tests that the cache is blindly being used
        obj._clear_caches()
        ret = obj.val1()
        assert "val1" in obj._cache
        assert ret == obj.ref1
        new = 77.0
        obj._fill_cache("val1", new)
        assert obj.val1() == new

    # Managed property
    def test_val2_lookup(self, obj):
        obj._clear_caches()
        assert "val2" not in obj._cache
        assert obj.val2 == obj.ref2
        ret = obj.val2
        assert "val2" in obj._cache
        assert obj._cache["val2"] == ret

    def test_val2_inject(self, obj):
        obj._clear_caches()
        ret = obj.val2
        assert "val2" in obj._cache
        assert ret == obj.ref2
        new = 77.0
        obj._fill_cache("val2", new)
        assert obj.val2 == new

        # Setter on cached attribute

    def test_val3_set(self, obj):
        obj._clear_caches()
        assert obj.val3 == obj.ref3
        new = 99.0
        obj.val3 = new
        assert obj.val3 == new
        assert obj._cache["val3"] == new

    def test_val3_del(self, obj):
        # Check that deleting the property removes it from cache,
        obj._clear_caches()
        assert obj.val3 == obj.ref3
        assert "val3" in obj._cache
        del obj.val3
        assert "val3" not in obj._cache
        # But allows it to work as usual afterwards
        assert obj.val3 == obj.ref3
        assert "val3" in obj._cache

    # Pass args
    def test_val4_args(self, obj):
        obj._clear_caches()
        assert obj.val4(1, 2) == 1 + 2 + obj.ref4
        # Further calls should yield the old result
        # this arguably shouldn't be cached...
        assert obj.val4(3, 4) == 1 + 2 + obj.ref4

    # Pass args and kwargs
    def test_val5_kwargs(self, obj):
        obj._clear_caches()
        assert obj.val5(5, s="abc") == 5 * "abc"

        assert obj.val5(5, s="!!!") == 5 * "abc"

    # property decorator, with universe validation
    def test_val6_universe_validation(self, obj):
        obj._clear_caches()
        assert not hasattr(obj, "_cache_key")
        assert "val6" not in obj._cache
        assert "val6" not in obj.universe._cache["_valid"]

        ret = obj.val6  # Trigger caching
        assert obj.val6 == obj.ref6
        assert ret is obj.val6
        assert "val6" in obj._cache
        assert "val6" in obj.universe._cache["_valid"]
        assert obj._cache_key in obj.universe._cache["_valid"]["val6"]
        assert obj._cache["val6"] is ret

        # Invalidate cache at universe level
        obj.universe._cache["_valid"]["val6"].clear()
        ret2 = obj.val6
        assert ret2 is obj.val6
        assert ret2 is not ret

        # Clear obj cache and access again
        obj._clear_caches()
        ret3 = obj.val6
        assert ret3 is obj.val6
        assert ret3 is not ret2
        assert ret3 is not ret


class TestConvFloat(object):
    @pytest.mark.parametrize(
        "s, output", [("0.45", 0.45), (".45", 0.45), ("a.b", "a.b")]
    )
    def test_float(self, s, output):
        assert util.conv_float(s) == output

    @pytest.mark.parametrize(
        "input, output",
        [
            (("0.45", "0.56", "6.7"), [0.45, 0.56, 6.7]),
            (("0.45", "a.b", "!!"), [0.45, "a.b", "!!"]),
        ],
    )
    def test_map(self, input, output):
        ret = [util.conv_float(el) for el in input]
        assert ret == output


class TestFixedwidthBins(object):
    def test_keys(self):
        ret = util.fixedwidth_bins(0.5, 1.0, 2.0)
        for k in ["Nbins", "delta", "min", "max"]:
            assert k in ret

    def test_ValueError(self):
        with pytest.raises(ValueError):
            util.fixedwidth_bins(0.1, 5.0, 4.0)

    @pytest.mark.parametrize(
        "delta, xmin, xmax, output_Nbins, output_delta, output_min, output_max",
        [
            (0.1, 4.0, 5.0, 10, 0.1, 4.0, 5.0),
            (0.4, 4.0, 5.0, 3, 0.4, 3.9, 5.1),
        ],
    )
    def test_usage(
        self,
        delta,
        xmin,
        xmax,
        output_Nbins,
        output_delta,
        output_min,
        output_max,
    ):
        ret = util.fixedwidth_bins(delta, xmin, xmax)
        assert ret["Nbins"] == output_Nbins
        assert ret["delta"] == output_delta
        assert ret["min"], output_min
        assert ret["max"], output_max


@pytest.fixture
def atoms():
    from MDAnalysisTests import make_Universe

    u = make_Universe(extras=("masses",), size=(3, 1, 1))
    return u.atoms


@pytest.mark.parametrize(
    "weights,result",
    [
        (None, None),
        ("mass", np.array([5.1, 4.2, 3.3])),
        (np.array([12.0, 1.0, 12.0]), np.array([12.0, 1.0, 12.0])),
        ([12.0, 1.0, 12.0], np.array([12.0, 1.0, 12.0])),
        (range(3), np.arange(3, dtype=int)),
    ],
)
def test_check_weights_ok(atoms, weights, result):
    assert_array_equal(util.get_weights(atoms, weights), result)


@pytest.mark.parametrize(
    "weights",
    [
        42,
        "geometry",
        np.array(1.0),
        np.array([12.0, 1.0, 12.0, 1.0]),
        [12.0, 1.0],
        np.array([[12.0, 1.0, 12.0]]),
        np.array([[12.0, 1.0, 12.0], [12.0, 1.0, 12.0]]),
    ],
)
def test_check_weights_raises_ValueError(atoms, weights):
    with pytest.raises(ValueError):
        util.get_weights(atoms, weights)


class TestGuessFormat(object):
    """Test guessing of format from filenames

    Tests also getting the appropriate Parser and Reader from a
    given filename
    """

    # list of known formats, followed by the desired Parser and Reader
    # None indicates that there isn't a Reader for this format
    # All formats call fallback to the MinimalParser
    formats = [
        (
            "CHAIN",
            mda.topology.MinimalParser.MinimalParser,
            mda.coordinates.chain.ChainReader,
        ),
        (
            "CONFIG",
            mda.topology.DLPolyParser.ConfigParser,
            mda.coordinates.DLPoly.ConfigReader,
        ),
        (
            "CRD",
            mda.topology.CRDParser.CRDParser,
            mda.coordinates.CRD.CRDReader,
        ),
        (
            "DATA",
            mda.topology.LAMMPSParser.DATAParser,
            mda.coordinates.LAMMPS.DATAReader,
        ),
        (
            "DCD",
            mda.topology.MinimalParser.MinimalParser,
            mda.coordinates.DCD.DCDReader,
        ),
        (
            "DMS",
            mda.topology.DMSParser.DMSParser,
            mda.coordinates.DMS.DMSReader,
        ),
        (
            "GMS",
            mda.topology.GMSParser.GMSParser,
            mda.coordinates.GMS.GMSReader,
        ),
        (
            "GRO",
            mda.topology.GROParser.GROParser,
            mda.coordinates.GRO.GROReader,
        ),
        (
            "HISTORY",
            mda.topology.DLPolyParser.HistoryParser,
            mda.coordinates.DLPoly.HistoryReader,
        ),
        (
            "INPCRD",
            mda.topology.MinimalParser.MinimalParser,
            mda.coordinates.INPCRD.INPReader,
        ),
        (
            "LAMMPS",
            mda.topology.MinimalParser.MinimalParser,
            mda.coordinates.LAMMPS.DCDReader,
        ),
        (
            "MDCRD",
            mda.topology.MinimalParser.MinimalParser,
            mda.coordinates.TRJ.TRJReader,
        ),
        (
            "MMTF",
            mda.topology.MMTFParser.MMTFParser,
            mda.coordinates.MMTF.MMTFReader,
        ),
        (
            "MOL2",
            mda.topology.MOL2Parser.MOL2Parser,
            mda.coordinates.MOL2.MOL2Reader,
        ),
        (
            "NC",
            mda.topology.MinimalParser.MinimalParser,
            mda.coordinates.TRJ.NCDFReader,
        ),
        (
            "NCDF",
            mda.topology.MinimalParser.MinimalParser,
            mda.coordinates.TRJ.NCDFReader,
        ),
        (
            "PDB",
            mda.topology.PDBParser.PDBParser,
            mda.coordinates.PDB.PDBReader,
        ),
        (
            "PDBQT",
            mda.topology.PDBQTParser.PDBQTParser,
            mda.coordinates.PDBQT.PDBQTReader,
        ),
        ("PRMTOP", mda.topology.TOPParser.TOPParser, None),
        (
            "PQR",
            mda.topology.PQRParser.PQRParser,
            mda.coordinates.PQR.PQRReader,
        ),
        ("PSF", mda.topology.PSFParser.PSFParser, None),
        (
            "RESTRT",
            mda.topology.MinimalParser.MinimalParser,
            mda.coordinates.INPCRD.INPReader,
        ),
        ("TOP", mda.topology.TOPParser.TOPParser, None),
        ("TPR", mda.topology.TPRParser.TPRParser, None),
        (
            "TRJ",
            mda.topology.MinimalParser.MinimalParser,
            mda.coordinates.TRJ.TRJReader,
        ),
        (
            "TRR",
            mda.topology.MinimalParser.MinimalParser,
            mda.coordinates.TRR.TRRReader,
        ),
        ("XML", mda.topology.HoomdXMLParser.HoomdXMLParser, None),
        (
            "XPDB",
            mda.topology.ExtendedPDBParser.ExtendedPDBParser,
            mda.coordinates.PDB.ExtendedPDBReader,
        ),
        (
            "XTC",
            mda.topology.MinimalParser.MinimalParser,
            mda.coordinates.XTC.XTCReader,
        ),
        (
            "XYZ",
            mda.topology.XYZParser.XYZParser,
            mda.coordinates.XYZ.XYZReader,
        ),
        (
            "TRZ",
            mda.topology.MinimalParser.MinimalParser,
            mda.coordinates.TRZ.TRZReader,
        ),
    ]
    # list of possible compressed extensions
    # include no extension too!
    compressed_extensions = [".bz2", ".gz"]

    @pytest.mark.parametrize(
        "extention",
        [format_tuple[0].upper() for format_tuple in formats]
        + [format_tuple[0].lower() for format_tuple in formats],
    )
    def test_get_extention(self, extention):
        """Check that get_ext works"""
        file_name = "file.{0}".format(extention)
        a, b = util.get_ext(file_name)

        assert a == "file"
        assert b == extention.lower()

    @pytest.mark.parametrize(
        "extention",
        [format_tuple[0].upper() for format_tuple in formats]
        + [format_tuple[0].lower() for format_tuple in formats],
    )
    def test_compressed_without_compression_extention(self, extention):
        """Check that format suffixed by compressed extension works"""
        file_name = "file.{0}".format(extention)
        a = util.format_from_filename_extension(file_name)
        # expect answer to always be uppercase
        assert a == extention.upper()

    @pytest.mark.parametrize(
        "extention",
        [format_tuple[0].upper() for format_tuple in formats]
        + [format_tuple[0].lower() for format_tuple in formats],
    )
    @pytest.mark.parametrize("compression_extention", compressed_extensions)
    def test_compressed(self, extention, compression_extention):
        """Check that format suffixed by compressed extension works"""
        file_name = "file.{0}{1}".format(extention, compression_extention)
        a = util.format_from_filename_extension(file_name)
        # expect answer to always be uppercase
        assert a == extention.upper()

    @pytest.mark.parametrize(
        "extention",
        [format_tuple[0].upper() for format_tuple in formats]
        + [format_tuple[0].lower() for format_tuple in formats],
    )
    def test_guess_format(self, extention):
        file_name = "file.{0}".format(extention)
        a = util.guess_format(file_name)
        # expect answer to always be uppercase
        assert a == extention.upper()

    @pytest.mark.parametrize(
        "extention",
        [format_tuple[0].upper() for format_tuple in formats]
        + [format_tuple[0].lower() for format_tuple in formats],
    )
    @pytest.mark.parametrize("compression_extention", compressed_extensions)
    def test_guess_format_compressed(self, extention, compression_extention):
        file_name = "file.{0}{1}".format(extention, compression_extention)
        a = util.guess_format(file_name)
        # expect answer to always be uppercase
        assert a == extention.upper()

    @pytest.mark.parametrize(
        "extention, parser",
        [
            (format_tuple[0], format_tuple[1])
            for format_tuple in formats
            if format_tuple[1] is not None
        ],
    )
    def test_get_parser(self, extention, parser):
        file_name = "file.{0}".format(extention)
        a = mda.topology.core.get_parser_for(file_name)

        assert a == parser

    @pytest.mark.parametrize(
        "extention, parser",
        [
            (format_tuple[0], format_tuple[1])
            for format_tuple in formats
            if format_tuple[1] is not None
        ],
    )
    @pytest.mark.parametrize("compression_extention", compressed_extensions)
    def test_get_parser_compressed(
        self, extention, parser, compression_extention
    ):
        file_name = "file.{0}{1}".format(extention, compression_extention)
        a = mda.topology.core.get_parser_for(file_name)

        assert a == parser

    @pytest.mark.parametrize(
        "extention",
        [
            (format_tuple[0], format_tuple[1])
            for format_tuple in formats
            if format_tuple[1] is None
        ],
    )
    def test_get_parser_invalid(self, extention):
        file_name = "file.{0}".format(extention)
        with pytest.raises(ValueError):
            mda.topology.core.get_parser_for(file_name)

    @pytest.mark.parametrize(
        "extention, reader",
        [
            (format_tuple[0], format_tuple[2])
            for format_tuple in formats
            if format_tuple[2] is not None
        ],
    )
    def test_get_reader(self, extention, reader):
        file_name = "file.{0}".format(extention)
        a = mda.coordinates.core.get_reader_for(file_name)

        assert a == reader

    @pytest.mark.parametrize(
        "extention, reader",
        [
            (format_tuple[0], format_tuple[2])
            for format_tuple in formats
            if format_tuple[2] is not None
        ],
    )
    @pytest.mark.parametrize("compression_extention", compressed_extensions)
    def test_get_reader_compressed(
        self, extention, reader, compression_extention
    ):
        file_name = "file.{0}{1}".format(extention, compression_extention)
        a = mda.coordinates.core.get_reader_for(file_name)

        assert a == reader

    @pytest.mark.parametrize(
        "extention",
        [
            (format_tuple[0], format_tuple[2])
            for format_tuple in formats
            if format_tuple[2] is None
        ],
    )
    def test_get_reader_invalid(self, extention):
        file_name = "file.{0}".format(extention)
        with pytest.raises(ValueError):
            mda.coordinates.core.get_reader_for(file_name)

    def test_check_compressed_format_TypeError(self):
        with pytest.raises(TypeError):
            util.check_compressed_format(1234, "bz2")

    def test_format_from_filename_TypeError(self):
        with pytest.raises(TypeError):
            util.format_from_filename_extension(1234)

    def test_guess_format_stream_ValueError(self):
        # This stream has no name, so can't guess format
        s = StringIO("this is a very fun file")
        with pytest.raises(ValueError):
            util.guess_format(s)

    def test_from_ndarray(self):
        fn = np.zeros((3, 3))
        rd = mda.coordinates.core.get_reader_for(fn)
        assert rd == mda.coordinates.memory.MemoryReader


class TestUniqueRows(object):
    def test_unique_rows_2(self):
        a = np.array([[0, 1], [1, 2], [2, 1], [0, 1], [0, 1], [2, 1]])

        assert_array_equal(
            util.unique_rows(a), np.array([[0, 1], [1, 2], [2, 1]])
        )

    def test_unique_rows_3(self):
        a = np.array([[0, 1, 2], [0, 1, 2], [2, 3, 4], [0, 1, 2]])

        assert_array_equal(
            util.unique_rows(a), np.array([[0, 1, 2], [2, 3, 4]])
        )

    def test_unique_rows_with_view(self):
        # unique_rows doesn't work when flags['OWNDATA'] is False,
        # happens when second dimension is created through broadcast
        a = np.array([1, 2])

        assert_array_equal(util.unique_rows(a[None, :]), np.array([[1, 2]]))


class TestGetWriterFor(object):
    def test_no_filename_argument(self):
        # Does ``get_writer_for`` fails as expected when provided no
        # filename arguments
        with pytest.raises(TypeError):
            mda.coordinates.core.get_writer_for()

    def test_precedence(self):
        writer = mda.coordinates.core.get_writer_for("test.pdb", "GRO")
        assert writer == mda.coordinates.GRO.GROWriter
        # Make sure ``get_writer_for`` uses *format* if provided

    def test_missing_extension(self):
        # Make sure ``get_writer_for`` behave as expected if *filename*
        # has no extension
        with pytest.raises(ValueError):
            mda.coordinates.core.get_writer_for(filename="test", format=None)

    def test_extension_empty_string(self):
        """
        Test format=''.

        Raises TypeError because format can be only None or
        valid formats.
        """
        with pytest.raises(ValueError):
            mda.coordinates.core.get_writer_for(filename="test", format="")

    def test_file_no_extension(self):
        """No format given"""
        with pytest.raises(ValueError):
            mda.coordinates.core.get_writer_for("outtraj")

    def test_wrong_format(self):
        # Make sure ``get_writer_for`` fails if the format is unknown
        with pytest.raises(TypeError):
            mda.coordinates.core.get_writer_for(
                filename="fail_me", format="UNK"
            )

    def test_compressed_extension(self):
        for ext in (".gz", ".bz2"):
            fn = "test.gro" + ext
            writer = mda.coordinates.core.get_writer_for(filename=fn)
            assert writer == mda.coordinates.GRO.GROWriter
            # Make sure ``get_writer_for`` works with compressed file file names

    def test_compressed_extension_fail(self):
        for ext in (".gz", ".bz2"):
            fn = "test.unk" + ext
            # Make sure ``get_writer_for`` fails if an unknown format is compressed
            with pytest.raises(TypeError):
                mda.coordinates.core.get_writer_for(filename=fn)

    def test_non_string_filename(self):
        # Does ``get_writer_for`` fails with non string filename, no format
        with pytest.raises(ValueError):
            mda.coordinates.core.get_writer_for(
                filename=StringIO(), format=None
            )

    def test_multiframe_failure(self):
        # does ``get_writer_for`` fail with invalid format and multiframe not None
        with pytest.raises(TypeError):
            mda.coordinates.core.get_writer_for(
                filename="fail_me", format="UNK", multiframe=True
            )
            mda.coordinates.core.get_writer_for(
                filename="fail_me", format="UNK", multiframe=False
            )

    def test_multiframe_nonsense(self):
        with pytest.raises(ValueError):
            mda.coordinates.core.get_writer_for(
                filename="this.gro", multiframe="sandwich"
            )

    formats = [
        # format name, related class, singleframe, multiframe
        ("CRD", mda.coordinates.CRD.CRDWriter, True, False),
        ("DATA", mda.coordinates.LAMMPS.DATAWriter, True, False),
        ("DCD", mda.coordinates.DCD.DCDWriter, True, True),
        # ('ENT', mda.coordinates.PDB.PDBWriter, True, False),
        ("GRO", mda.coordinates.GRO.GROWriter, True, False),
        ("LAMMPS", mda.coordinates.LAMMPS.DCDWriter, True, True),
        ("MOL2", mda.coordinates.MOL2.MOL2Writer, True, True),
        ("NCDF", mda.coordinates.TRJ.NCDFWriter, True, True),
        ("NULL", mda.coordinates.null.NullWriter, True, True),
        # ('PDB', mda.coordinates.PDB.PDBWriter, True, True), special case, done separately
        ("PDBQT", mda.coordinates.PDBQT.PDBQTWriter, True, False),
        ("PQR", mda.coordinates.PQR.PQRWriter, True, False),
        ("TRR", mda.coordinates.TRR.TRRWriter, True, True),
        ("XTC", mda.coordinates.XTC.XTCWriter, True, True),
        ("XYZ", mda.coordinates.XYZ.XYZWriter, True, True),
        ("TRZ", mda.coordinates.TRZ.TRZWriter, True, True),
    ]

    @pytest.mark.parametrize(
        "format, writer",
        [
            (format_tuple[0], format_tuple[1])
            for format_tuple in formats
            if format_tuple[2] is True
        ],
    )
    def test_singleframe(self, format, writer):
        assert (
            mda.coordinates.core.get_writer_for(
                "this", format=format, multiframe=False
            )
            == writer
        )

    @pytest.mark.parametrize(
        "format",
        [
            (format_tuple[0], format_tuple[1])
            for format_tuple in formats
            if format_tuple[2] is False
        ],
    )
    def test_singleframe_fails(self, format):
        with pytest.raises(TypeError):
            mda.coordinates.core.get_writer_for(
                "this", format=format, multiframe=False
            )

    @pytest.mark.parametrize(
        "format, writer",
        [
            (format_tuple[0], format_tuple[1])
            for format_tuple in formats
            if format_tuple[3] is True
        ],
    )
    def test_multiframe(self, format, writer):
        assert (
            mda.coordinates.core.get_writer_for(
                "this", format=format, multiframe=True
            )
            == writer
        )

    @pytest.mark.parametrize(
        "format",
        [
            format_tuple[0]
            for format_tuple in formats
            if format_tuple[3] is False
        ],
    )
    def test_multiframe_fails(self, format):
        with pytest.raises(TypeError):
            mda.coordinates.core.get_writer_for(
                "this", format=format, multiframe=True
            )

    def test_get_writer_for_pdb(self):
        assert (
            mda.coordinates.core.get_writer_for(
                "this", format="PDB", multiframe=False
            )
            == mda.coordinates.PDB.PDBWriter
        )
        assert (
            mda.coordinates.core.get_writer_for(
                "this", format="PDB", multiframe=True
            )
            == mda.coordinates.PDB.MultiPDBWriter
        )
        assert (
            mda.coordinates.core.get_writer_for(
                "this", format="ENT", multiframe=False
            )
            == mda.coordinates.PDB.PDBWriter
        )
        assert (
            mda.coordinates.core.get_writer_for(
                "this", format="ENT", multiframe=True
            )
            == mda.coordinates.PDB.MultiPDBWriter
        )


class TestBlocksOf(object):
    def test_blocks_of_1(self):
        arr = np.arange(16).reshape(4, 4)

        view = util.blocks_of(arr, 1, 1)

        assert view.shape == (4, 1, 1)
        assert_array_almost_equal(
            view, np.array([[[0]], [[5]], [[10]], [[15]]])
        )

        # Change my view, check changes are reflected in arr
        view[:] = 1001

        assert_array_almost_equal(
            arr,
            np.array(
                [
                    [1001, 1, 2, 3],
                    [4, 1001, 6, 7],
                    [8, 9, 1001, 11],
                    [12, 13, 14, 1001],
                ]
            ),
        )

    def test_blocks_of_2(self):
        arr = np.arange(16).reshape(4, 4)

        view = util.blocks_of(arr, 2, 2)

        assert view.shape == (2, 2, 2)
        assert_array_almost_equal(
            view, np.array([[[0, 1], [4, 5]], [[10, 11], [14, 15]]])
        )

        view[0] = 100
        view[1] = 200

        assert_array_almost_equal(
            arr,
            np.array(
                [
                    [100, 100, 2, 3],
                    [100, 100, 6, 7],
                    [8, 9, 200, 200],
                    [12, 13, 200, 200],
                ]
            ),
        )

    def test_blocks_of_3(self):
        # testing non square array
        arr = np.arange(32).reshape(8, 4)

        view = util.blocks_of(arr, 2, 1)

        assert view.shape == (4, 2, 1)

    def test_blocks_of_4(self):
        # testing block exceeding array size results in empty view
        arr = np.arange(4).reshape(2, 2)
        view = util.blocks_of(arr, 3, 3)
        assert view.shape == (0, 3, 3)
        view[:] = 100
        assert_array_equal(arr, np.arange(4).reshape(2, 2))

    def test_blocks_of_ValueError(self):
        arr = np.arange(16).reshape(4, 4)
        with pytest.raises(ValueError):
            util.blocks_of(arr, 2, 1)  # blocks don't fit
        with pytest.raises(ValueError):
            util.blocks_of(arr[:, ::2], 2, 1)  # non-contiguous input


@pytest.mark.parametrize(
    "arr,answer",
    [
        ([2, 3, 4, 7, 8, 9, 10, 15, 16], [[2, 3, 4], [7, 8, 9, 10], [15, 16]]),
        ([11, 12, 13, 14, 15, 16], [[11, 12, 13, 14, 15, 16]]),
        ([1, 2, 2, 2, 3, 6], [[1, 2, 2, 2, 3], [6]]),
    ],
)
def test_group_same_or_consecutive_integers(arr, answer):
    assert_equal(util.group_same_or_consecutive_integers(arr), answer)


class TestNamespace(object):
    @staticmethod
    @pytest.fixture()
    def ns():
        return util.Namespace()

    def test_getitem(self, ns):
        ns.this = 42
        assert ns["this"] == 42

    def test_getitem_KeyError(self, ns):
        with pytest.raises(KeyError):
            dict.__getitem__(ns, "this")

    def test_setitem(self, ns):
        ns["this"] = 42

        assert ns["this"] == 42

    def test_delitem(self, ns):
        ns["this"] = 42
        assert "this" in ns
        del ns["this"]
        assert "this" not in ns

    def test_delitem_AttributeError(self, ns):
        with pytest.raises(AttributeError):
            del ns.this

    def test_setattr(self, ns):
        ns.this = 42

        assert ns.this == 42

    def test_getattr(self, ns):
        ns["this"] = 42

        assert ns.this == 42

    def test_getattr_AttributeError(self, ns):
        with pytest.raises(AttributeError):
            getattr(ns, "this")

    def test_delattr(self, ns):
        ns["this"] = 42

        assert "this" in ns
        del ns.this
        assert "this" not in ns

    def test_eq(self, ns):
        ns["this"] = 42

        ns2 = util.Namespace()
        ns2["this"] = 42

        assert ns == ns2

    def test_len(self, ns):
        assert len(ns) == 0
        ns["this"] = 1
        ns["that"] = 2
        assert len(ns) == 2

    def test_iter(self, ns):
        ns["this"] = 12
        ns["that"] = 24
        ns["other"] = 48

        seen = []
        for val in ns:
            seen.append(val)
        for val in ["this", "that", "other"]:
            assert val in seen


class TestTruncateInteger(object):
    @pytest.mark.parametrize(
        "a, b",
        [
            ((1234, 1), 4),
            ((1234, 2), 34),
            ((1234, 3), 234),
            ((1234, 4), 1234),
            ((1234, 5), 1234),
        ],
    )
    def test_ltruncate_int(self, a, b):
        assert util.ltruncate_int(*a) == b


class TestFlattenDict(object):
    def test_flatten_dict(self):
        d = {
            "A": {1: ("a", "b", "c")},
            "B": {2: ("c", "d", "e")},
            "C": {3: ("f", "g", "h")},
        }
        result = util.flatten_dict(d)

        for k in result:
            assert type(k) == tuple
            assert len(k) == 2
            assert k[0] in d
            assert k[1] in d[k[0]]
            assert result[k] in d[k[0]].values()


class TestStaticVariables(object):
    """Tests concerning the decorator @static_variables"""

    def test_static_variables(self):
        x = [0]

        @static_variables(foo=0, bar={"test": x})
        def myfunc():
            assert myfunc.foo == 0
            assert type(myfunc.bar) is type(dict())
            if "test2" not in myfunc.bar:
                myfunc.bar["test2"] = "a"
            else:
                myfunc.bar["test2"] += "a"
            myfunc.bar["test"][0] += 1
            return myfunc.bar["test"]

        assert hasattr(myfunc, "foo")
        assert hasattr(myfunc, "bar")

        y = myfunc()
        assert y is x
        assert x[0] == 1
        assert myfunc.bar["test"][0] == 1
        assert myfunc.bar["test2"] == "a"

        x = [0]
        y = myfunc()
        assert y is not x
        assert myfunc.bar["test"][0] == 2
        assert myfunc.bar["test2"] == "aa"


class TestWarnIfNotUnique(object):
    """Tests concerning the decorator @warn_if_not_unique"""

    def warn_msg(self, func, group, group_name):
        msg = (
            "{}.{}(): {} {} contains duplicates. Results might be "
            "biased!".format(
                group.__class__.__name__,
                func.__name__,
                group_name,
                group.__repr__(),
            )
        )
        return msg

    def test_warn_if_not_unique(self, atoms):
        # Check that the warn_if_not_unique decorator has a "static variable"
        # warn_if_not_unique.warned:
        assert hasattr(warn_if_not_unique, "warned")
        assert warn_if_not_unique.warned is False

    def test_warn_if_not_unique_once_outer(self, atoms):

        # Construct a scenario with two nested functions, each one decorated
        # with @warn_if_not_unique:

        @warn_if_not_unique
        def inner(group):
            if not group.isunique:
                # The inner function should not trigger a warning, and the state
                # of warn_if_not_unique.warned should reflect that:
                assert warn_if_not_unique.warned is True
            return 0

        @warn_if_not_unique
        def outer(group):
            return inner(group)

        # Check that no warning is raised for a unique group:
        assert atoms.isunique

        with warnings.catch_warnings():
            warnings.simplefilter("error")
            x = outer(atoms)
            assert x == 0

        # Check that a warning is raised for a group with duplicates:
        ag = atoms + atoms[0]
        msg = self.warn_msg(outer, ag, "'ag'")
        with pytest.warns(DuplicateWarning) as w:
            assert warn_if_not_unique.warned is False
            x = outer(ag)
            # Assert that the "warned" state is restored:
            assert warn_if_not_unique.warned is False
            # Check correct function execution:
            assert x == 0
            # Only one warning must have been raised:
            assert len(w) == 1
            # For whatever reason pytest.warns(DuplicateWarning, match=msg)
            # doesn't work, so we compare the recorded warning message instead:
            assert w[0].message.args[0] == msg
            # Make sure the warning uses the correct stacklevel and references
            # this file instead of MDAnalysis/lib/util.py:
            assert w[0].filename == __file__

    def test_warned_state_restored_on_failure(self, atoms):

        # A decorated function raising an exception:
        @warn_if_not_unique
        def thisfails(group):
            raise ValueError()

        ag = atoms + atoms[0]
        msg = self.warn_msg(thisfails, ag, "'ag'")
        with pytest.warns(DuplicateWarning) as w:
            assert warn_if_not_unique.warned is False
            with pytest.raises(ValueError):
                thisfails(ag)
            # Assert that the "warned" state is restored despite `thisfails`
            # raising an exception:
            assert warn_if_not_unique.warned is False
            assert len(w) == 1
            assert w[0].message.args[0] == msg
            assert w[0].filename == __file__

    def test_warn_if_not_unique_once_inner(self, atoms):

        # Construct a scenario with two nested functions, each one decorated
        # with @warn_if_not_unique, but the outer function adds a duplicate
        # to the group:

        @warn_if_not_unique
        def inner(group):
            return 0

        @warn_if_not_unique
        def outer(group):
            dupgroup = group + group[0]
            return inner(dupgroup)

        # Check that even though outer() is called the warning is raised for
        # inner():
        msg = self.warn_msg(inner, atoms + atoms[0], "'dupgroup'")
        with pytest.warns(DuplicateWarning) as w:
            assert warn_if_not_unique.warned is False
            x = outer(atoms)
            # Assert that the "warned" state is restored:
            assert warn_if_not_unique.warned is False
            # Check correct function execution:
            assert x == 0
            # Only one warning must have been raised:
            assert len(w) == 1
            assert w[0].message.args[0] == msg
            assert w[0].filename == __file__

    def test_warn_if_not_unique_multiple_references(self, atoms):
        ag = atoms + atoms[0]
        aag = ag
        aaag = aag

        @warn_if_not_unique
        def func(group):
            return group.isunique

        # Check that the warning message contains the names of all references to
        # the group in alphabetic order:
        msg = self.warn_msg(func, ag, "'aaag' a.k.a. 'aag' a.k.a. 'ag'")
        with pytest.warns(DuplicateWarning) as w:
            x = func(ag)
            # Assert that the "warned" state is restored:
            assert warn_if_not_unique.warned is False
            # Check correct function execution:
            assert x is False
            # Check warning message:
            assert w[0].message.args[0] == msg
            # Check correct file referenced:
            assert w[0].filename == __file__

    def test_warn_if_not_unique_unnamed(self, atoms):

        @warn_if_not_unique
        def func(group):
            pass

        msg = self.warn_msg(
            func,
            atoms + atoms[0],
            "'unnamed {}'".format(atoms.__class__.__name__),
        )
        with pytest.warns(DuplicateWarning) as w:
            func(atoms + atoms[0])
            # Check warning message:
            assert w[0].message.args[0] == msg

    def test_warn_if_not_unique_fails_for_non_groupmethods(self):

        @warn_if_not_unique
        def func(group):
            pass

        class dummy(object):
            pass

        with pytest.raises(AttributeError):
            func(dummy())

    def test_filter_duplicate_with_userwarning(self, atoms):

        @warn_if_not_unique
        def func(group):
            pass

        with warnings.catch_warnings(record=True) as record:
            warnings.resetwarnings()
            warnings.filterwarnings("ignore", category=UserWarning)
            with warnings.catch_warnings():
                warnings.simplefilter("error")
                func(atoms)
            assert len(record) == 0


class TestCheckCoords(object):
    """Tests concerning the decorator @check_coords"""

    prec = 6

    def test_default_options(self):
        a_in = np.zeros(3, dtype=np.float32)
        b_in = np.ones(3, dtype=np.float32)
        b_in2 = np.ones((2, 3), dtype=np.float32)

        @check_coords("a", "b")
        def func(a, b):
            # check that enforce_copy is True by default:
            assert a is not a_in
            assert b is not b_in
            # check that convert_single is True by default:
            assert a.shape == (1, 3)
            assert b.shape == (1, 3)
            return a + b

        # check that allow_single is True by default:
        res = func(a_in, b_in)
        # check that reduce_result_if_single is True by default:
        assert res.shape == (3,)
        # check correct function execution:
        assert_array_equal(res, b_in)

        # check that check_lenghts_match is True by default:
        with pytest.raises(ValueError):
            res = func(a_in, b_in2)

    @pytest.fixture()
    def atomgroup(self):
        u = mda.Universe(PSF, DCD)
        return u.atoms

    # check atomgroup handling with every option except allow_atomgroup
    @pytest.mark.parametrize("enforce_copy", [True, False])
    @pytest.mark.parametrize("enforce_dtype", [True, False])
    @pytest.mark.parametrize("allow_single", [True, False])
    @pytest.mark.parametrize("convert_single", [True, False])
    @pytest.mark.parametrize("reduce_result_if_single", [True, False])
    @pytest.mark.parametrize("check_lengths_match", [True, False])
    def test_atomgroup(
        self,
        atomgroup,
        enforce_copy,
        enforce_dtype,
        allow_single,
        convert_single,
        reduce_result_if_single,
        check_lengths_match,
    ):
        ag1 = atomgroup
        ag2 = atomgroup

        @check_coords(
            "ag1",
            "ag2",
            enforce_copy=enforce_copy,
            enforce_dtype=enforce_dtype,
            allow_single=allow_single,
            convert_single=convert_single,
            reduce_result_if_single=reduce_result_if_single,
            check_lengths_match=check_lengths_match,
            allow_atomgroup=True,
        )
        def func(ag1, ag2):
            assert_allclose(ag1, ag2)
            assert isinstance(ag1, np.ndarray)
            assert isinstance(ag2, np.ndarray)
            assert ag1.dtype == ag2.dtype == np.float32
            return ag1 + ag2

        res = func(ag1, ag2)

        assert_allclose(res, atomgroup.positions * 2)

    def test_atomgroup_not_allowed(self, atomgroup):

        @check_coords("ag1", allow_atomgroup=False)
        def func(ag1):
            return ag1

        with pytest.raises(TypeError, match="allow_atomgroup is False"):
            _ = func(atomgroup)

    def test_atomgroup_not_allowed_default(self, atomgroup):

        @check_coords("ag1")
        def func_default(ag1):
            return ag1

        with pytest.raises(TypeError, match="allow_atomgroup is False"):
            _ = func_default(atomgroup)

    def test_enforce_copy(self):

        a_2d = np.ones((1, 3), dtype=np.float32)
        b_1d = np.zeros(3, dtype=np.float32)
        c_2d = np.zeros((1, 6), dtype=np.float32)[:, ::2]
        d_2d = np.zeros((1, 3), dtype=np.int64)

        @check_coords("a", "b", "c", "d", enforce_copy=False)
        def func(a, b, c, d):
            # Assert that if enforce_copy is False:
            # no copy is made if input shape, order, and dtype are correct:
            assert a is a_2d
            # a copy is made if input shape has to be changed:
            assert b is not b_1d
            # a copy is made if input order has to be changed:
            assert c is not c_2d
            # a copy is made if input dtype has to be changed:
            assert d is not d_2d
            # Assert correct dtype conversion:
            assert d.dtype == np.float32
            assert_almost_equal(d, d_2d, self.prec)
            # Assert all shapes are converted to (1, 3):
            assert a.shape == b.shape == c.shape == d.shape == (1, 3)
            return a + b + c + d

        # Call func() to:
        # - test the above assertions
        # - ensure that input of single coordinates is simultaneously possible
        #   with different shapes (3,) and (1, 3)
        res = func(a_2d, b_1d, c_2d, d_2d)
        # Since some inputs are not 1d, even though reduce_result_if_single is
        # True, the result must have shape (1, 3):
        assert res.shape == (1, 3)
        # check correct function execution:
        assert_array_equal(res, a_2d)

    def test_no_allow_single(self):

        @check_coords("a", allow_single=False)
        def func(a):
            pass

        with pytest.raises(ValueError) as err:
            func(np.zeros(3, dtype=np.float32))
            assert err.msg == ("func(): a.shape must be (n, 3), got (3,).")

    def test_no_convert_single(self):

        a_1d = np.arange(-3, 0, dtype=np.float32)

        @check_coords("a", enforce_copy=False, convert_single=False)
        def func(a):
            # assert no conversion and no copy were performed:
            assert a is a_1d
            return a

        res = func(a_1d)
        # Assert result has been reduced:
        assert res == a_1d[0]
        assert type(res) is np.float32

    def test_no_reduce_result_if_single(self):

        a_1d = np.zeros(3, dtype=np.float32)

        # Test without shape conversion:
        @check_coords(
            "a",
            enforce_copy=False,
            convert_single=False,
            reduce_result_if_single=False,
        )
        def func(a):
            return a

        res = func(a_1d)
        # make sure the input array is just passed through:
        assert res is a_1d

        # Test with shape conversion:
        @check_coords("a", enforce_copy=False, reduce_result_if_single=False)
        def func(a):
            return a

        res = func(a_1d)
        assert res.shape == (1, 3)
        assert_array_equal(res[0], a_1d)

    def test_no_check_lengths_match(self):

        a_2d = np.zeros((1, 3), dtype=np.float32)
        b_2d = np.zeros((3, 3), dtype=np.float32)

        @check_coords("a", "b", enforce_copy=False, check_lengths_match=False)
        def func(a, b):
            return a, b

        res_a, res_b = func(a_2d, b_2d)
        # Assert arrays are just passed through:
        assert res_a is a_2d
        assert res_b is b_2d

    def test_atomgroup_mismatched_lengths(self):
        u = mda.Universe(PSF, DCD)
        ag1 = u.select_atoms("index 0 to 10")
        ag2 = u.atoms

        @check_coords(
            "ag1", "ag2", check_lengths_match=True, allow_atomgroup=True
        )
        def func(ag1, ag2):

            return ag1, ag2

        with pytest.raises(
            ValueError, match="must contain the same number of " "coordinates"
        ):
            _, _ = func(ag1, ag2)

    def test_invalid_input(self):

        a_inv_dtype = np.array([["hello", "world", "!"]])
        a_inv_type = [[0.0, 0.0, 0.0]]
        a_inv_shape_1d = np.zeros(6, dtype=np.float32)
        a_inv_shape_2d = np.zeros((3, 2), dtype=np.float32)

        @check_coords("a")
        def func(a):
            pass

        with pytest.raises(TypeError) as err:
            func(a_inv_dtype)
            assert err.msg.startswith(
                "func(): a.dtype must be convertible to " "float32, got "
            )

        with pytest.raises(TypeError) as err:
            func(a_inv_type)
            assert err.msg == (
                "func(): Parameter 'a' must be a numpy.ndarray, "
                "got <class 'list'>."
            )

        with pytest.raises(ValueError) as err:
            func(a_inv_shape_1d)
            assert err.msg == (
                "func(): a.shape must be (3,) or (n, 3), got " "(6,)."
            )

        with pytest.raises(ValueError) as err:
            func(a_inv_shape_2d)
            assert err.msg == (
                "func(): a.shape must be (3,) or (n, 3), got " "(3, 2)."
            )

    def test_usage_with_kwargs(self):

        a_2d = np.zeros((1, 3), dtype=np.float32)

        @check_coords("a", enforce_copy=False)
        def func(a, b, c=0):
            return a, b, c

        # check correct functionality if passed as keyword argument:
        a, b, c = func(a=a_2d, b=0, c=1)
        assert a is a_2d
        assert b == 0
        assert c == 1

    def test_wrong_func_call(self):

        @check_coords("a", enforce_copy=False)
        def func(a, b, c=0):
            pass

        # Make sure invalid call marker is present:
        func._invalid_call = False

        # usage with posarg doubly defined:
        assert not func._invalid_call
        with pytest.raises(TypeError):
            func(0, a=0)  # pylint: disable=redundant-keyword-arg
        assert func._invalid_call
        func._invalid_call = False

        # usage with missing posargs:
        assert not func._invalid_call
        with pytest.raises(TypeError):
            func(0)
        assert func._invalid_call
        func._invalid_call = False

        # usage with missing posargs (supplied as kwargs):
        assert not func._invalid_call
        with pytest.raises(TypeError):
            func(a=0, c=1)
        assert func._invalid_call
        func._invalid_call = False

        # usage with too many posargs:
        assert not func._invalid_call
        with pytest.raises(TypeError):
            func(0, 0, 0, 0)
        assert func._invalid_call
        func._invalid_call = False

        # usage with unexpected kwarg:
        assert not func._invalid_call
        with pytest.raises(TypeError):
            func(a=0, b=0, c=1, d=1)  # pylint: disable=unexpected-keyword-arg
        assert func._invalid_call
        func._invalid_call = False

    def test_wrong_decorator_usage(self):

        # usage without parantheses:
        @check_coords
        def func():
            pass

        with pytest.raises(TypeError):
            func()

        # usage without arguments:
        with pytest.raises(ValueError) as err:

            @check_coords()
            def func():
                pass

            assert err.msg == (
                "Decorator check_coords() cannot be used "
                "without positional arguments."
            )

        # usage with defaultarg:
        with pytest.raises(ValueError) as err:

            @check_coords("a")
            def func(a=1):
                pass

            assert err.msg == (
                "In decorator check_coords(): Name 'a' doesn't "
                "correspond to any positional argument of the "
                "decorated function func()."
            )

        # usage with invalid parameter name:
        with pytest.raises(ValueError) as err:

            @check_coords("b")
            def func(a):
                pass

            assert err.msg == (
                "In decorator check_coords(): Name 'b' doesn't "
                "correspond to any positional argument of the "
                "decorated function func()."
            )


@pytest.mark.parametrize("old_name", (None, "MDAnalysis.Universe"))
@pytest.mark.parametrize("new_name", (None, "Multiverse"))
@pytest.mark.parametrize("remove", (None, "99.0.0", 2099))
@pytest.mark.parametrize("message", (None, "use the new stuff"))
def test_deprecate(old_name, new_name, remove, message, release="2.7.1"):
    def AlternateUniverse(anything):
        # important: first line needs to be """\ so that textwrap.dedent()
        # works
        """\
        AlternateUniverse provides a true view of the Universe.

        Parameters
        ----------
        anything : object

        Returns
        -------
        truth

        """
        return True

    oldfunc = util.deprecate(
        AlternateUniverse,
        old_name=old_name,
        new_name=new_name,
        release=release,
        remove=remove,
        message=message,
    )
    # match_expr changed to match (Issue 2329)
    with pytest.warns(DeprecationWarning, match="`.+` is deprecated"):
        oldfunc(42)

    doc = oldfunc.__doc__
    name = old_name if old_name else AlternateUniverse.__name__

    deprecation_line_1 = ".. deprecated:: {0}".format(release)
    assert re.search(deprecation_line_1, doc)

    if message:
        deprecation_line_2 = message
    else:
        if new_name is None:
            default_message = "`{0}` is deprecated!".format(name)
        else:
            default_message = "`{0}` is deprecated, use `{1}` instead!".format(
                name, new_name
            )
        deprecation_line_2 = default_message
    assert re.search(deprecation_line_2, doc)

    if remove:
        deprecation_line_3 = "`{0}` will be removed in release {1}".format(
            name, remove
        )
        assert re.search(deprecation_line_3, doc)

    # check that the old docs are still present
    assert re.search(textwrap.dedent(AlternateUniverse.__doc__), doc)


def test_deprecate_missing_release_ValueError():
    with pytest.raises(ValueError):
        util.deprecate(mda.Universe)


def test_set_function_name(name="bar"):
    def foo():
        pass

    util._set_function_name(foo, name)
    assert foo.__name__ == name


@pytest.mark.parametrize(
    "text",
    (
        "",
        "one line text",
        "  one line with leading space",
        "multiline\n\n   with some\n   leading space",
        "   multiline\n\n   with all\n   leading space",
    ),
)
def test_dedent_docstring(text):
    doc = util.dedent_docstring(text)
    for line in doc.splitlines():
        assert line == line.lstrip()


class TestCheckBox(object):

    prec = 6
    ref_ortho = np.ones(3, dtype=np.float32)
    ref_tri_vecs = np.array(
        [[1, 0, 0], [0, 1, 0], [0, 2**0.5, 2**0.5]], dtype=np.float32
    )

    @pytest.mark.parametrize(
        "box",
        (
            [1, 1, 1, 90, 90, 90],
            (1, 1, 1, 90, 90, 90),
            ["1", "1", 1, 90, "90", "90"],
            ("1", "1", 1, 90, "90", "90"),
            np.array(["1", "1", 1, 90, "90", "90"]),
            np.array([1, 1, 1, 90, 90, 90], dtype=np.float32),
            np.array([1, 1, 1, 90, 90, 90], dtype=np.float64),
            np.array(
                [1, 1, 1, 1, 1, 1, 90, 90, 90, 90, 90, 90], dtype=np.float32
            )[::2],
        ),
    )
    def test_check_box_ortho(self, box):
        boxtype, checked_box = util.check_box(box)
        assert boxtype == "ortho"
        assert_allclose(checked_box, self.ref_ortho)
        assert checked_box.dtype == np.float32
        assert checked_box.flags["C_CONTIGUOUS"]

    def test_check_box_None(self):
        with pytest.raises(ValueError, match="Box is None"):
            util.check_box(None)

    @pytest.mark.parametrize(
        "box",
        (
            [1, 1, 2, 45, 90, 90],
            (1, 1, 2, 45, 90, 90),
            ["1", "1", 2, 45, "90", "90"],
            ("1", "1", 2, 45, "90", "90"),
            np.array(["1", "1", 2, 45, "90", "90"]),
            np.array([1, 1, 2, 45, 90, 90], dtype=np.float32),
            np.array([1, 1, 2, 45, 90, 90], dtype=np.float64),
            np.array(
                [1, 1, 1, 1, 2, 2, 45, 45, 90, 90, 90, 90], dtype=np.float32
            )[::2],
        ),
    )
    def test_check_box_tri_vecs(self, box):
        boxtype, checked_box = util.check_box(box)
        assert boxtype == "tri_vecs"
        assert_almost_equal(checked_box, self.ref_tri_vecs, self.prec)
        assert checked_box.dtype == np.float32
        assert checked_box.flags["C_CONTIGUOUS"]

    def test_check_box_wrong_data(self):
        with pytest.raises(ValueError):
            wrongbox = ["invalid", 1, 1, 90, 90, 90]
            boxtype, checked_box = util.check_box(wrongbox)

    def test_check_box_wrong_shape(self):
        with pytest.raises(ValueError):
            wrongbox = np.ones((3, 3), dtype=np.float32)
            boxtype, checked_box = util.check_box(wrongbox)


class StoredClass:
    """
    A simple class that takes positional and keyword arguments of various types
    """

    @store_init_arguments
    def __init__(self, a, b, /, *args, c="foo", d="bar", e="foobar", **kwargs):
        self.a = a
        self.b = b
        self.c = c
        self.d = d
        self.e = e
        self.args = args
        self.kwargs = kwargs

    def copy(self):
        kwargs = copy.deepcopy(self._kwargs)
        args = kwargs.pop("args", tuple())
        new = self.__class__(kwargs.pop("a"), kwargs.pop("b"), *args, **kwargs)
        return new


class TestStoreInitArguments:
    def test_store_arguments_default(self):
        store = StoredClass("parsnips", ["roast"])
        assert store.a == store._kwargs["a"] == "parsnips"
        assert store.b is store._kwargs["b"] == ["roast"]
        assert store._kwargs["c"] == "foo"
        assert store._kwargs["d"] == "bar"
        assert store._kwargs["e"] == "foobar"
        assert "args" not in store._kwargs.keys()
        assert "kwargs" not in store._kwargs.keys()
        assert store.args is ()

        store2 = store.copy()
        assert store2.__dict__ == store.__dict__
        assert store2.__dict__["b"] is not store.__dict__["b"]

    def test_store_arguments_withkwargs(self):
        store = StoredClass(
            "parsnips",
            "roast",
            "honey",
            "glaze",
            c="richard",
            d="has",
            e="a",
            f="recipe",
            g="allegedly",
        )
        assert store.a == store._kwargs["a"] == "parsnips"
        assert store.b == store._kwargs["b"] == "roast"
        assert store.c == store._kwargs["c"] == "richard"
        assert store.d == store._kwargs["d"] == "has"
        assert store.e == store._kwargs["e"] == "a"
        assert store.kwargs["f"] == store._kwargs["f"] == "recipe"
        assert store.kwargs["g"] == store._kwargs["g"] == "allegedly"
        assert store.args[0] == store._kwargs["args"][0] == "honey"
        assert store.args[1] == store._kwargs["args"][1] == "glaze"

        store2 = store.copy()
        assert store2.__dict__ == store.__dict__


@pytest.mark.xfail(
    os.name == "nt", reason="util.which does not get right binary on Windows"
)
def test_which():
    wmsg = "This method is deprecated"

    with pytest.warns(DeprecationWarning, match=wmsg):
        assert util.which("python") == shutil.which("python")


@pytest.mark.parametrize(
    "modulename,is_installed",
    [("math", True), ("sys", True), ("some_weird_module_name", False)],
)
def test_is_installed(modulename, is_installed):
    assert util.is_installed(modulename) == is_installed
